# fetch_files tab for bfit
# Derek Fujimoto
# Nov 2017

from tkinter import *
from tkinter import ttk, messagebox, filedialog
from bfit import logger_name
import bdata as bd
from bdata import bdata, bmerged
from functools import partial
from bfit.backend.fitdata import fitdata
from bfit.backend.entry_color_set import on_focusout, on_entry_click
import bfit.backend.colors as colors
from bfit.global_variables import KEYVARS
from bfit.gui.dataline import dataline
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import datetime, os, logging, glob, re

__doc__="""
    """

# =========================================================================== #
class fetch_files(object):
    """
        Data fields:
            
            asym_type: StringVar, drawing style
            bfit: pointer to parent class
            base_bins = IntVar, number of bins to use as baseline on scan ends
            canvas_frame_id: id number of frame in canvas
            check_rebin: IntVar for handling rebin aspect of checkall
            check_bin_remove: StringVar for handing omission of 1F data
            check_state: BooleanVar for handling check all
            check_state_data: BooleanVar for handling check_all_data
            check_state_fit: BooleanVar for handling check_all_fit
            check_state_res: BooleanVar for handling check_all_res
            data_canvas: canvas object allowing for scrolling 
            dataline_frame: frame holding all the data lines. Exists as a window
                in the data_canvas
            data_lines: dictionary of dataline obj, keyed by run number
            data_lines_old: dictionary of removed dataline obj, keyed by run number
            entry_asym_type: combobox for asym calc and draw type
            entry_run: entry to put in run number string
            fet_entry_frame: frame of fetch tab
            filter_opt: StringVar, holds state of filter radio buttons
            listbox_history: listbox for run input history
            max_number_fetched: max number of files you can fetch
            omit_state: BooleanVar, if true set omit all final incomplete scans
            run: StringVar input to fetch runs.
            runmode_label: display run mode
            runmode: display run mode list of strings
            text_filter: Text box for filter commands input
            year: IntVar of year to fetch runs from 
    """
    
    runmode_relabel = {'20':'SLR (20)', 
                       '1f':'Frequency Scan (1f/1x)', 
                       '1x':'Frequency Scan (1f/1x)', 
                       '1w':'Frequency Comb (1w)', 
                       '2e':'Random Freq. (2e)', 
                       '1n':'Rb Cell Scan (1n)', 
                       '1e':'Field Scan (1e)', 
                       '2h':'Alpha Tagged (2h)'}
                       
    equivalent_modes = [['1f', '1x'], ['20', '2h']]
                       
    run_number_starter_line = '40001 40002+40003 40005-40010 (run numbers)'
    bin_remove_starter_line = '24 100-200 (bins)'
    max_number_fetched = 500
    nhistory = 10
    
    # ======================================================================= #
    def __init__(self, fetch_data_tab, bfit):
        
        # get logger
        self.logger = logging.getLogger(logger_name)
        self.logger.debug('Initializing')
    
        # initialize
        self.bfit = bfit
        self.data_lines = {}
        self.data_lines_old = {}
        self.fit_input_tabs = {}
        self.check_rebin = IntVar()
        self.check_bin_remove = StringVar()
        self.check_state = BooleanVar()
        self.fetch_data_tab = fetch_data_tab
        
        # Frame for specifying files -----------------------------------------
        fet_entry_frame = ttk.Labelframe(fetch_data_tab, text='Specify Files')
        self.year = IntVar()
        self.run = StringVar()
        
        self.year.set(self.bfit.get_latest_year())
        
        entry_year = Spinbox(fet_entry_frame, textvariable=self.year, width=5, 
                             from_=2000, to=datetime.datetime.today().year)
        entry_run = Entry(fet_entry_frame, textvariable=self.run, width=85)
        entry_run.insert(0, self.run_number_starter_line)
        entry_run.config(foreground=colors.entry_grey)
        
        # history list
        self.listbox_history = Listbox(fetch_data_tab, selectmode=SINGLE)
        entry_fn = partial(on_entry_click, text=self.run_number_starter_line, \
                            entry=entry_run)
        on_focusout_fn = partial(on_focusout, text=self.run_number_starter_line, \
                            entry=entry_run)
        entry_run.bind('<FocusIn>', entry_fn)
        entry_run.bind('<FocusOut>', on_focusout_fn)
        entry_run.bind('<Leave>', lambda event: self.history_hide(event, 'entry'))
        
        entry_run.bind('<Enter>', self.history_show)
        self.listbox_history.bind('<Leave>', lambda event: self.history_hide(event, 'history'))
        self.listbox_history.bind("<<ListboxSelect>>", self.history_set)
        
        # fetch button
        fetch = ttk.Button(fet_entry_frame, text='Fetch', command=self.get_data)
        update = ttk.Button(fet_entry_frame, text='Update', command=self.update_data)
        
        # grid and labels
        fet_entry_frame.grid(column=0, row=0, sticky=(N, W, E), columnspan=2, padx=5, pady=5)
        ttk.Label(fet_entry_frame, text="Year:").grid(column=0, row=0, sticky=W)
        entry_year.grid(column=1, row=0, sticky=(W))
        ttk.Label(fet_entry_frame, text="Run Number:").grid(column=2, row=0, sticky=W)
        entry_run.grid(column=3, row=0, sticky=W)
        fetch.grid(column=4, row=0, sticky=E)
        update.grid(column=5, row=0, sticky=E)
        self.listbox_history.grid(column=3, row=1, sticky=W)
        
        # padding 
        for child in fet_entry_frame.winfo_children(): 
            child.grid_configure(padx=5, pady=5)
        
        self.listbox_history.grid_forget()
        
        # Frame for run mode -------------------------------------------------
        runmode_label_frame = ttk.Labelframe(fetch_data_tab, pad=(10, 5, 10, 5), \
                text='Run Mode', )
        
        self.runmode_label = ttk.Label(runmode_label_frame, text="", justify=CENTER)
        
        # Scrolling frame to hold datalines
        yscrollbar = ttk.Scrollbar(fetch_data_tab, orient=VERTICAL)         
        self.data_canvas = Canvas(fetch_data_tab, bd=0,              # make a canvas for scrolling
                yscrollcommand=yscrollbar.set,                      # scroll command receive
                scrollregion=(0, 0, 5000, 5000), confine=True)       # default size
        yscrollbar.config(command=self.data_canvas.yview)           # scroll command send
        dataline_frame = ttk.Frame(self.data_canvas, pad=5)          # holds 
        
        self.canvas_frame_id = self.data_canvas.create_window((0, 0),    # make window which can scroll
                window=dataline_frame, 
                anchor='nw')
        dataline_frame.bind("<Configure>", self.config_canvas) # bind resize to alter scrollable region
        self.data_canvas.bind("<Configure>", self.config_dataline_frame) # bind resize to change size of contained frame
        
        # Frame to hold everything on the right ------------------------------
        bigright_frame = ttk.Frame(fetch_data_tab, pad=5)
        
        # asymmetry calculation
        style_frame = ttk.Labelframe(bigright_frame, text='Asymmetry Calculation', \
                pad=5)
        self.asym_type = StringVar()
        self.asym_type.set('')
        self.entry_asym_type = ttk.Combobox(style_frame, \
                textvariable=self.asym_type, state='readonly', \
                width=20)
        self.entry_asym_type['values'] = ()
        
        style_frame.grid(column=0, row=0, sticky=(W, N, E))
        style_frame.grid_columnconfigure(0, weight=1)
        self.entry_asym_type.grid(column=0, row=0, sticky=(N, E, W), padx=10)
        
        # Frame for group set options ----------------------------------------
        right_frame = ttk.Labelframe(bigright_frame, \
                text='Operations on Active Items', pad=15)
        
        check_remove = ttk.Button(right_frame, text='Remove', \
                command=self.remove_all, pad=5)
        check_draw = ttk.Button(right_frame, text='Draw', \
                command=lambda:self.draw_all('data'), pad=5)
        
        
        # Scan repair options ------------------------------------------------
        frame_scan = ttk.Labelframe(right_frame, \
                text='Scan Repair Options', pad=5)
        
        check_rebin_label = ttk.Label(frame_scan, text="Rebin:", pad=5)
        check_rebin_box = Spinbox(frame_scan, from_=1, to=100, width=3, \
                textvariable=self.check_rebin, 
                command=lambda: self.set_all('rebin'))
        
        check_bin_remove_label = ttk.Label(frame_scan, text="Remove:", pad=5)
        check_bin_remove_entry = Entry(frame_scan, \
                textvariable=self.check_bin_remove, width=20)
        
        self.omit_state = BooleanVar()
        omit_checkbox = ttk.Checkbutton(frame_scan, 
                text='Omit incomplete final scan', variable=self.omit_state, 
                onvalue=True, offvalue=False, pad=5, 
                command=lambda: self.set_all('omit_scan'))
        self.omit_state.set(False)
        
        self.base_bins = IntVar()
        base_label = ttk.Label(frame_scan, text="baseline bins", pad=5)
        base_spinbox = Spinbox(frame_scan, from_=0, to=1000, width=4, \
                               textvariable=self.base_bins, 
                               command=lambda: self.set_all('base_bins'))
        self.base_bins.set(0)
        
        # grid
        check_rebin_label.grid(     column=0, row=0, sticky='e')
        check_rebin_box.grid(       column=1, row=0, sticky='w')
        check_bin_remove_label.grid(column=0, row=1, sticky='e')
        check_bin_remove_entry.grid(column=1, row=1, sticky='w', columnspan=2)
        omit_checkbox.grid(column=0, row=2, sticky='w', columnspan=2)
        base_spinbox.grid(column=0, row=3, sticky='e')
        base_label.grid(column=1, row=3, sticky='w')
        
        # key bindings
        check_bin_remove_entry.bind('<KeyRelease>', lambda x: self.set_all('omit'))
        check_rebin_box.bind('<KeyRelease>', lambda x: self.set_all('rebin'))
        base_spinbox.bind('<KeyRelease>', lambda x: self.set_all('base_bins'))
        
        # checkboxes
        right_checkbox_frame = ttk.Frame(right_frame)
        
        check_all_box = ttk.Checkbutton(right_checkbox_frame, 
                text='State', variable=self.check_state, 
                onvalue=True, offvalue=False, pad=5, command=self.check_all)
        self.check_state.set(True)
        
        self.check_state_data = BooleanVar()        
        check_data_box = ttk.Checkbutton(right_checkbox_frame, 
                text='Data', variable=self.check_state_data, 
                onvalue=True, offvalue=False, pad=5, command=self.check_all_data)
        self.check_state_data.set(True)
        
        self.check_state_fit = BooleanVar()        
        check_fit_box = ttk.Checkbutton(right_checkbox_frame, 
                text='Fit', variable=self.check_state_fit, 
                onvalue=True, offvalue=False, pad=5, command=self.check_all_fit)
        
        self.check_state_res = BooleanVar()        
        check_res_box = ttk.Checkbutton(right_checkbox_frame, 
                text='Res', variable=self.check_state_res, 
                onvalue=True, offvalue=False, pad=5, command=self.check_all_res)
                
        check_toggle_button = ttk.Button(right_frame, \
                text='Toggle All Check States', command=self.toggle_all, pad=5)
        
        # add grey to check_bin_remove_entry
        check_bin_remove_entry.insert(0, self.bin_remove_starter_line)
        
        check_entry_fn = partial(on_entry_click, \
                text=self.bin_remove_starter_line, \
                entry=check_bin_remove_entry)
        
        check_on_focusout_fn = partial(on_focusout, \
                text=self.bin_remove_starter_line, \
                entry=check_bin_remove_entry)
        
        check_bin_remove_entry.bind('<FocusIn>', check_entry_fn)
        check_bin_remove_entry.bind('<FocusOut>', check_on_focusout_fn)
        check_bin_remove_entry.config(foreground=colors.entry_grey)
                    
        # grid
        runmode_label_frame.grid(column=2, row=0, sticky=(N, W, E, S), pady=5, padx=5)
        self.runmode_label.grid(column=0, row=0, sticky=(N, W, E))
        
        bigright_frame.grid(column=2, row=1, rowspan=2, sticky='new')
        
        self.data_canvas.grid(column=0, row=1, sticky=(E, W, S, N), padx=5, pady=5)
        yscrollbar.grid(column=1, row=1, sticky=(W, S, N), pady=5)
        
        check_all_box.grid(        column=0, row=0, sticky=(N))
        check_data_box.grid(        column=1, row=0, sticky=(N))
        check_fit_box.grid(         column=2, row=0, sticky=(N))
        check_res_box.grid(         column=3, row=0, sticky=(N)) 
        
        right_frame.grid(           column=0, row=1, sticky=(N, E, W), pady=5)
        r = 0
        right_checkbox_frame.grid(  column=0, row=r, sticky=(N), columnspan=2); r+= 1
        check_toggle_button.grid(   column=0, row=r, sticky=(N, E, W), columnspan=2, pady=1, padx=5); r+= 1
        check_draw.grid(            column=0, row=r, sticky=(N, W, E), pady=5, padx=5);
        check_remove.grid(          column=1, row=r, sticky=(N, E, W), pady=5, padx=5); r+= 1
        frame_scan.grid(            column=0, row=r, sticky=(N, E, W), pady=5, padx=5, columnspan=2); r+= 1        
        
        # filtering -----------------------------------------------------------
        frame_filter = ttk.Labelframe(bigright_frame, text='Filter Runs', pad=5)
        button_filter_instr = ttk.Button(frame_filter, text='Instructions', 
                                         command=self.show_filter_instructions)
        self.text_filter = Text(frame_filter, width=35, height=5, state='normal', 
                                wrap='none')
        
        frame_filter_button = ttk.Frame(frame_filter, pad=5)
        button_filter = ttk.Button(frame_filter_button, text='Filter', 
                                         command=self.filter_runs)
        self.filter_opt = StringVar()
        self.filter_opt.set('activate')
        radio_filter_activate = ttk.Radiobutton(frame_filter_button, 
                                       variable=self.filter_opt, 
                                       value='activate',
                                       text='By deactivation',
                                       )
        radio_filter_remove = ttk.Radiobutton(frame_filter_button, 
                                       variable=self.filter_opt, 
                                       value='remove',
                                       text='By removal',
                                       )
        
        button_filter.grid(column=0, row=0, rowspan=2, sticky='w', padx=5)
        radio_filter_activate.grid(column=1, row=0, sticky='w', padx=5)
        radio_filter_remove.grid(column=1, row=1, sticky='w', padx=5)
        
        # final grid
        frame_filter.grid(column=0, row=2, sticky='new', pady=5)
        r = 0
        button_filter_instr.grid(column=0, row=r, sticky='new', pady=5); r+= 1
        self.text_filter.grid(column=0, row=r, sticky='new', pady=5); r+= 1
        frame_filter_button.grid(column=0, row=r, sticky='new', pady=5); r+= 1
        
        # resizing
        fetch_data_tab.grid_columnconfigure(0, weight=1)        # main area
        fetch_data_tab.grid_rowconfigure(1, weight=1)            # main area
        
        for i in range(3):
            if i%2 == 0:    fet_entry_frame.grid_columnconfigure(i, weight=2)
        fet_entry_frame.grid_columnconfigure(3, weight=1)
            
        self.data_canvas.grid_columnconfigure(0, weight=1)    # fetch frame 
        self.data_canvas.grid_rowconfigure(0, weight=1)
            
        # passing
        self.entry_run = entry_run
        self.entry_year = entry_year
        self.check_rebin_box = check_rebin_box
        self.check_bin_remove_entry = check_bin_remove_entry
        self.check_all_box = check_all_box
        self.dataline_frame = dataline_frame

        self.logger.debug('Initialization success.')
    
    # ======================================================================= #
    def __del__(self):
        
        # delete lists and dictionaries
        if hasattr(self, 'data_lines'):      del self.data_lines
        if hasattr(self, 'data_lines_old'):  del self.data_lines_old
        
        # kill buttons and frame
        try:
            for child in self.fetch_data_tab.winfo_children():
                child.destroy()
            self.fetch_data_tab.destroy()
        except Exception:
            pass
        
    # ======================================================================= #
    def _check_condition(self, data, string):
        """Check whether a run satisfies the condition given as a string"""
        
        # determine what variable is being tested
        keyvars = np.array(list(KEYVARS.keys()))
        keylens = [len(k) for k in keyvars]
        keyvars = keyvars[np.argsort(keylens)[::-1]]
        
        found = False
        for key in keyvars:
            if key in string:
                found = True
                break
        
        if not found:
            msg = 'Variable not recognized in "%s"' % string
            messagebox.showerror('Filter', msg)
            raise RuntimeError(msg)
        
        # replace key with value
        val, err = data.get_values(KEYVARS[key])
        string = string.replace(key, str(val))
        
        # check for nan
        if 'nan' in string:
            string = string.replace('nan', 'np.nan')
        
        # evaluate
        return eval(string)
    
    # ======================================================================= #
    def _do_check_all(self, state, var, box):
        """
            Force all tickboxes of a given type to be in a given state, assuming 
            the tickbox is active
        """
        
        self.logger.info('Changing state of all %s tickboxes to %s', var, state)
        for dline in self.data_lines.values():
            
            # check if line is selected
            if not dline.check_state.get() and var != 'check_state':
                continue
            
            # check if tickbox is object variable
            if hasattr(dline, box):
                
                # check if tickbox is disabled
                if str(getattr(dline, box)['state']) == 'disabled':
                    continue
                    
            # set value
            getattr(dline, var).set(state)
            
            # run check for other items
            if hasattr(dline, 'do_'+var):
                getattr(dline, 'do_'+var)()
    
    # ======================================================================= #
    def canvas_scroll(self, event):
        """Scroll canvas with files selected."""
        if event.num == 4:
            self.data_canvas.yview_scroll(-1, "units")
        elif event.num == 5:
            self.data_canvas.yview_scroll(1, "units")
    
    # ======================================================================= #
    def check_all(self):  
        self._do_check_all(self.check_state.get(), 'check_state', 'None')
        
    def check_all_data(self):  
        self._do_check_all(self.check_state_data.get(), 'check_data', 'None')
        
    def check_all_fit(self):  
        self._do_check_all(self.check_state_fit.get(), 'check_fit', 'draw_fit_checkbox')
    
    def check_all_res(self):  
        self._do_check_all(self.check_state_res.get(), 'check_res', 'draw_res_checkbox')    
        
    # ======================================================================= #
    def config_canvas(self, event):
        """Alter scrollable region based on canvas bounding box size. 
        (changes scrollbar properties)"""
        self.data_canvas.configure(scrollregion=self.data_canvas.bbox("all"))
    
    # ======================================================================= #
    def config_dataline_frame(self, event):
        """Alter size of contained frame in canvas. Allows for inside window to 
        be resized with mouse drag""" 
        self.data_canvas.itemconfig(self.canvas_frame_id, width=event.width)
        
    # ======================================================================= #
    def draw_all(self, figstyle, ignore_check=False):
        """
            Draw all data in data lines
            
            figstyle: one of "data", "fit", or "param" to choose which figure 
                    to draw in
            ignore_check: draw all with no regard to whether the run has been 
                    selected
        """
        
        self.logger.debug('Drawing all data (ignore check: %s)', ignore_check)
        
        # condense drawing into a funtion
        def draw_lines():
            for k, line in self.data_lines.items():
                if line.check_state.get() or ignore_check:
                    line.draw(figstyle)
                
        # get draw style
        style = self.bfit.draw_style.get()
        self.logger.debug('Draw style: "%s"',  style)
        
        # make new figure, draw stacked
        if style == 'stack':
            draw_lines()
            
        # overdraw in current figure, stacked
        elif style == 'redraw':
            self.bfit.draw_style.set('stack')
            
            if self.bfit.plt.plots[figstyle]:
                self.bfit.plt.clf(figstyle)
            
            draw_lines()
            self.bfit.draw_style.set('redraw')
            
        # make new figure, draw single
        elif style == 'new':
            self.bfit.draw_style.set('stack')
            self.bfit.plt.figure(figstyle)
            draw_lines()
            self.bfit.draw_style.set('new')
        else:
            s = "Draw style not recognized"
            messagebox.showerror(message=s)
            self.logger.error(s)
            raise ValueError(s)
    
        # remove legned if too many drawn values
        n_selected = sum([d.check_state.get() for d in self.data_lines.values()])
        if n_selected > self.bfit.legend_max_draw:
            
            try:
                self.bfit.plt.gca(figstyle).get_legend().remove()
            except AttributeError:
                pass
            else:
                self.bfit.plt.tight_layout(figstyle)
        
    # ======================================================================= #
    def export(self, directory=None):
        """Export all data files as csv"""
        
        # filename
        filename = self.bfit.fileviewer.default_export_filename
        if not filename: return
        self.logger.info('Exporting to file %s', filename)
        
        if directory is None:
            try:
                filename = os.path.join(filedialog.askdirectory(), filename)
            except TypeError:
                pass
        else:
            filename = os.path.join(directory, filename)
        
        # get data and write
        for k in self.bfit.data.keys():
            d = self.bfit.data[k]
            omit = d.omit.get()
            if omit == self.bin_remove_starter_line:    omit = ''
            self.bfit.export(d, filename%(d.year, d.run), rebin=d.rebin.get(), 
                             omit=omit)
        self.logger.debug('Success.')
        
    # ======================================================================= #
    def filter_runs(self):
        """Remove or select runs based on filter conditions"""
        
        # parse input string
        string = self.text_filter.get('1.0', END)
        lines = [l for l in string.split('\n') if l]
        
        # iterate on fetched data
        data_keep = []
        
        for k, d in self.bfit.data.items():
            
            # check conditions
            satisfy = all([self._check_condition(d, line) for line in lines])
            if satisfy:
                data_keep.append(k)
            
        # do filtering by selection
        if self.filter_opt.get() == 'activate':
            for k, d in self.bfit.data.items():
                d.check_state.set(k in data_keep)
                
        elif self.filter_opt.get() == 'remove':
            keys = tuple(self.data_lines.keys())
            for k in keys:
                if k not in data_keep:
                    self.data_lines[k].degrid()
    
    # ======================================================================= #
    def get_data(self):
        """Split data into parts, and assign to dictionary."""
        
        self.logger.debug('Fetching runs')
    
        # make list of run numbers, replace possible deliminators
        try:
            run_numbers = self.string2run(self.run.get())
        except ValueError:
            self.logger.exception('Bad run number string')
            raise ValueError('Bad run number string')
            
        # get list of merged runs
        merged_runs = self.get_merged_runs(self.run.get())
            
        # get the selected year
        year = int(self.year.get())
        
        # get data
        all_data = {}
        s = ['Failed to open run']
        for r in run_numbers:
            
            # read from archive
            try:
                all_data[r] = bdata(r, year=year)
            except (RuntimeError, ValueError):
                s.append("%d (%d)" % (r, year))
                    
        # print error message
        if len(s)>1:
            s = '\n'.join(s)
            print(s)
            self.logger.warning(s)
            messagebox.showinfo(message=s)
        
        # merge runs
        new_data = []
        for merge in merged_runs: 
            
            # collect data
            dat_to_merge = []
            for r in merge: 
                dat_to_merge.append(all_data.pop(r))
            
            # make bjoined object
            new_data.append(bmerged(dat_to_merge))
            
        new_data.extend(list(all_data.values()))
        
        # update object data lists
        data = {}
        for new_dat in new_data:
            
            # get key for data storage
            runkey = self.bfit.get_run_key(new_dat)

            # update data
            if runkey in self.bfit.data.keys():
                self.bfit.data[runkey].read()
                
            # new data
            else:
                data[runkey] = fitdata(self.bfit, new_dat)
    
        # check that data is all the same runtype
        run_types = [d.mode for d in self.bfit.data.values()]
        run_types = run_types + [d.mode for d in data.values()]
        
        # check for equivalent run modes
        selected_mode = [run_types[0]]
        for equiv in self.equivalent_modes:
            if selected_mode[0] in equiv:
                selected_mode = equiv
                break
        
        # different run types: select all runs of same type
        if not all([r in selected_mode for r in run_types]):
            
            # unique run modes
            run_type_unique = np.unique(run_types)
            
            # message for multiple run modes
            message = "Multiple run types detected:\n("
            message += ', '.join(run_type_unique)
            message += ')\n\nSelecting ' + ' and '.join(selected_mode) + ' runs.'
            messagebox.showinfo(message=message)
            
        # get only run_types[0]
        self.logger.debug('Fetching runs of mode %s', selected_mode)
        for k in tuple(data.keys()):
            if data[k].mode in selected_mode:
                self.bfit.data[k] = data[k]
            else:
                del data[k]
        
        # get runmode
        try:
            self.runmode = selected_mode
        except IndexError:
            s = 'No valid runs detected.'
            messagebox.showerror(message=s)
            self.logger.warning(s)
            raise RuntimeError(s)
            
        self.runmode_label['text'] = self.runmode_relabel[self.runmode[0]]
        
        # get area
        area = [d.area for d in self.bfit.data.values()]
        area = ''.join(np.unique(area))
        
        # set asym type comboboxes
        self.bfit.set_asym_calc_mode_box(self.runmode[0], self, area)
        self.bfit.set_asym_calc_mode_box(self.runmode[0], self.bfit.fit_files, area)
        
        keys_list = list(self.bfit.data.keys())
        keys_list.sort()
        
        # make lines
        n = 1
        for r in keys_list:
            
            # new line
            if r not in self.data_lines.keys():
                
                if r in self.data_lines_old.keys():
                    self.data_lines[r] = self.data_lines_old[r]
                    self.data_lines[r].set_label()
                    del self.data_lines_old[r]
                else:
                    self.data_lines[r] = dataline(\
                                            bfit = self.bfit, \
                                            lines_list = self.data_lines, \
                                            lines_list_old = self.data_lines_old, 
                                            fetch_tab_frame = self.dataline_frame, \
                                            bdfit = self.bfit.data[r], \
                                            row = n)
            self.data_lines[r].grid(n)
            n+=1
            
        # remove old runs, modes not selected
        for r in tuple(self.data_lines.keys()):
            if self.data_lines[r].bdfit.mode not in self.runmode:
                self.data_lines[r].degrid()
            
        # set nbm variable
        self.set_nbm()
        
        self.logger.info('Fetched runs %s', list(data.keys()))
        
    # ======================================================================= #
    def get_merged_runs(self, string):
        """
            Parse string, return list of lists of run numbers corresponding to 
            the data to merge
        """
        
        # find plus locations
        idx_plus = [m.start() for m in re.finditer(r'\+', string)]

        # remove spaces around plus
        for i in idx_plus[::-1]:
            if string[i+1] == ' ':
                string = string[:i+1]+string[i+2:]
            if string[i-1] == ' ':
                string = string[:i-1]+string[i:]
            
        # clean input
        string = re.sub('[, ;]', ' ', string)

        # add brackets
        string = ' '.join(['[%s]' % s if '+' in s else s for s in string.split()])

        # remove plusses
        string = re.sub(r'\+', ' ', string)

        # get strings for merging
        merge_strings = [*re.findall(r'\[(.*?)\]', string), 
                         *re.findall(r'\((.*?)\)', string), 
                         *re.findall(r'\{(.*?)\}', string)]

        
        # split up the input string by brackets
        merge_strings = [*re.findall(r'\[(.*?)\]', string), 
                         *re.findall(r'\((.*?)\)', string), 
                         *re.findall(r'\{(.*?)\}', string)]

        # get the run numbers in each of the merged string inputs
        merge_runs = [self.string2run(s) for s in merge_strings]
        
        return merge_runs
        
    # ======================================================================= #
    def history_set(self, x):
        """
            Set the history entry as the current fetch string
        """
        
        idx = self.listbox_history.curselection() 
        if not idx:
            on_focusout_fn(0)
            return 
            
        item = self.listbox_history.get(idx)
        self.listbox_history.delete(idx)
        self.listbox_history.insert(0, item)
        self.entry_run.delete(0, END)
        self.entry_run.insert(0, item)
        self.history_hide()
        self.entry_run.focus_set()
        
    # ======================================================================= #
    def history_show(self, x=None):
        """
            Show list widget with input history
        """
        if self.listbox_history.size()>0:
            self.listbox_history.config(height=self.listbox_history.size())
            x = self.entry_run.winfo_x()
            y = self.entry_run.winfo_y()
            dx = self.entry_run.winfo_width()
            self.listbox_history.place(x=x+6, y=y+30, width=dx)
            self.listbox_history.lift()
            
    # ======================================================================= #
    def history_hide(self, event=None, obj=None):
        """
            Hide list widget with input history
        """

        # don't hide if mousing over history
        if obj == 'entry' and 0 < event.x < self.entry_run.winfo_reqwidth() and event.y > 0:
            return  
        
        # hide the history box    
        self.listbox_history.place_forget()
        string = self.run.get() 
        
        if string != self.listbox_history.get(0) and \
           string != self.run_number_starter_line:
            self.listbox_history.insert(0, string)
            
        if self.listbox_history.size()>self.nhistory:
            self.listbox_history.delete(END)            
    
    # ======================================================================= #
    def remove_all(self):
        """Remove all data files from self.data_lines"""
        
        self.logger.info('Removing all data files')
        del_list = []
        for r in self.data_lines.keys():
            if self.data_lines[r].check_state.get():
                del_list.append(self.data_lines[r])
        
        for d in del_list:
            d.degrid()
    
    # ======================================================================= #
    def return_binder(self):
        """Switch between various functions of the enter button. """
        
        # check where the focus is
        focus_id = self.bfit.root.focus_get()
        
        # run or year entry
        if focus_id in [self.entry_run, self.entry_year]:
            self.logger.debug('Focus is: run or year entry')
            self.get_data()
            self.history_hide()
        
        # check all box
        elif focus_id == self.check_all_box:
            self.logger.debug('Focus is: check all box')
            self.draw_all()
            
        # filter
        elif focus_id == self.text_filter:
            self.logger.debug('Focus is: text_filter')
            self.filter_runs()
            
        else:
            pass

    # ======================================================================= #
    def set_all(self, changed, *args):
        """Set a particular property for all checked items. """
        
        self.logger.info('Set all')
        data = self.bfit.data
        
        # check all file lines
        for dat in data.values():
            
            # if checked
            if dat.check_state.get():
                
                # rebin
                if changed == 'rebin':
                    try:
                        rebin = self.check_rebin.get()
                    except TclError:
                        rebin = 1
                        
                    dat.rebin.set(rebin)
                
                # omit bins
                elif changed == 'omit':
                    dat.omit.set(self.check_bin_remove.get())
                
                # omit scan
                elif changed == 'omit_scan':
                    dat.omit_scan.set(self.omit_state.get())
                
                # baseline corrections
                elif changed == 'base_bins':
                    try:
                        val = self.base_bins.get()
                    except TclError:
                        val = 1
                    dat.base_bins.set(val)
                    
    # ======================================================================= #
    def set_all_labels(self):
        """Set lable text in all items """
        
        self.logger.info('Setting all label text')
        
        # check all file lines
        for r in self.data_lines.keys():
            self.data_lines[r].set_label()
            
        for r in self.data_lines_old.keys():
            self.data_lines_old[r].set_label()

    # ======================================================================= #
    def set_nbm(self):
        """
            Set the nbm variable based on the run mode
        """
        
        # check if data
        if not hasattr(self, 'runmode'):
            return
        
        # check run mode
        mode = self.runmode[0]
        
        self.bfit.set_nbm(mode)
    
    # ======================================================================= #
    def show_filter_instructions(self):
        """Display popup with instructions on filtering"""
        
        # instructions
        msg = 'Write one boolean expression \nper line using the below '+\
              'variables. \n\n'+\
              'Runs which satisfy all expressions \nare kept/activated\n\n'
              
        # variables
        keys = sorted(KEYVARS.keys())
        values = [KEYVARS[k] for k in keys]
        max_keylen = max(tuple(map(len, keys)))
        
        lines = [k.ljust(max_keylen+2) + v for k, v in zip(keys, values)]
              
        msg += '     ' + ('\n     '.join(lines))
        
        # example
        ex = 'ex: "10 < BIAS < 15"'
        msg += '\n\n' + ex
        
        # make window
        win = Toplevel(self.bfit.mainframe)
        win.title('How to filter runs')
        self.bfit.set_icon(win)
        
        frame = ttk.Frame(win, pad=10, relief='sunken')
        frame.grid(column=0, row=0)
        ttk.Label(frame, text=msg, justify='left').grid(column=0, row=0, pady=5)
        ttk.Button(frame, text='Close', 
                   command=lambda: win.destroy()).grid(column=0, row=1, pady=5)
            
    # ======================================================================= #
    def string2run(self, string):
        """Parse string, return list of run numbers"""
        
        # standardize deliminators
        full_string = re.sub(r'[, ;+\[\]\(\)\{\}]', ' ', string)
        full_string = full_string.replace(':', '-')
        part_string = full_string.split()
        
        # get list of run numbers
        run_numbers = []
        for s in part_string:
            
            # select a range
            if '-' in s:
                
                # get runs in range
                try:
                    spl = s.split('-')
                    
                    # look for "range to current run"
                    if spl[1] == '':
                        
                        # look for latest run by run number
                        if int(spl[0]) < 45000:
                            try:
                                dirloc = os.environ[self.bfit.bnmr_archive_label]
                            except KeyError:
                                dirloc = os.path.join(bd._mud_data, 'bnmr')
                        else:
                            try:
                                dirloc = os.environ[self.bfit.bnqr_archive_label]
                            except KeyError:
                                dirloc = os.path.join(bd._mud_data, 'bnqr')
                            
                        runlist = glob.glob(os.path.join(dirloc, 
                                                         str(self.year.get()), 
                                                         '*.msr'))
                        spl[1] = max([int(os.path.splitext(os.path.basename(r))[0]) 
                                   for r in runlist])
                        
                    rn_lims = tuple(map(int, spl))
                    
                # get bad range first value
                except ValueError:
                    run_numbers.append(int(s.replace('-', '')))
                
                # get range of runs
                else:
                    rns = np.arange(rn_lims[0], rn_lims[1]+1).tolist()
                    run_numbers.extend(rns)
            
            # get single run
            else:
                run_numbers.append(int(s))
        
        # sort
        run_numbers.sort()
        self.logger.debug('Parsed "%s" to run numbers (len: %d) %s', string, 
                          len(run_numbers), run_numbers)
        
        if len(run_numbers) > self.max_number_fetched:
            raise RuntimeWarning("Too many files selected (max %d)." \
                                                    % self.max_number_fetched)
        return run_numbers
    
    # ======================================================================= #
    def toggle_all(self):
        """Toggle all tickboxes"""
        self.logger.info('Toggling all tickboxes')
        for k in self.data_lines.keys():
            state = not self.data_lines[k].check_state.get()
            self.data_lines[k].check_state.set(state)

    # ======================================================================= #
    def update_data(self):
        """
            Re-fetch all fetched runs.
            Update labels
        """
        
        self.logger.info('Updating data')
        
        # fetch
        for dat in self.bfit.data.values():
            dat.read()
            
        # update text
        for line in self.data_lines.values():
            line.set_check_text()
            line.update_label()



# Search for runs based on some parameters
# Derek Fujimoto
# Sep 2024

from bdata import bdata
import numpy as np
import mudpy as mp
import pandas as pd

class searchable_bdata(bdata):

    def get_attribute(self, parname):
        """Get attribute of the bdata object to search

        Args:
            parname (str): name of parameter
        Returns:
            float|str: corresponding to that parameters value
        """

        # get value
        val = getattr(self, parname)

        if type(val) is mp.mvar.mvar:
            return val.mean
        else:
            return val

class database(object):

    def __init__(self, runs=None, years=None):
        """Create object with searchable parameters in order to find runs

        Args:
            runs (iterable): list of runs to look into
            years (iterable): list of years to look into
        """

        # get the data
        if runs is not None and years is not None:
            self.data = []
            for r in runs:
                for y in years:
                    try:
                        self.data.append(searchable_bdata(r, y))
                    except RuntimeError:
                        pass
            self.data = np.array(self.data)

    def get_runs(self):
        """Return a dataframe of runs and years for all entries in the database"""

        runyear = np.zeros((len(self.data), 2), dtype=int)

        for i, d in enumerate(self.data):
            runyear[i, 0] = d.run
            runyear[i, 1] = d.year
        return runyear
        # return pd.DataFrame({'run':runyear[:,0], 'year':runyear[:,1]}, dtype=int)

    def print_par(self, parname):
        for d in self.data:
            val = getattr(d, parname)
            print(f'{d.year}  {d.run}  {val}')

    def search(self, parname, searchterm):
        """Do the search.

        Args:
            parname (str): name of parameter to search through
            low (float): parameter must be larger than low
            high (float): parameter must be smaller than high
            substr (str): for string searches, this string must be in parameter

        Returns:
            A copy of this object with a smaller database size
        """

        # get the search term
        low = high = substr = None

        if type(searchterm) is str:
            substr = searchterm

        elif type(searchterm) in (np.ndarray, list, tuple) and len(searchterm) == 2:
            low, high = searchterm

        else:
            raise RuntimeError('bad input')

        # get the parameter
        values = np.array([b.get_attribute(parname) for b in self.data])

        # index of those that satisfy
        idx = np.full(len(values), True)
        if low is not None:
            idx *= values > low

        if high is not None:
            idx *= values < high

        if substr is not None:
            idx = np.array([substr in v for v in values])

        # get those values
        data = self.data[idx]

        # make new object
        obj = database()
        obj.data = data

        return obj



# test inspect tab
# Derek Fujimoto
# Feb 2021

from numpy.testing import *
import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
import bdata as bd
from bfit import pulsed_exp, minuit
from bfit.fitting.leastsquares import LeastSquares
from bfit.gui.bfit import bfit

# filter unneeded warnings
import pytest
pytestmark = pytest.mark.filterwarnings('ignore:2020')

def with_bfit(function):
    
    def wrapper(*args, **kwargs):
        # make gui
        b = bfit(None, True)
        tab = b.fit_files
        tab2 = b.fetch_files
        b.notebook.select(2)
        
        tab2.year.set(2020)
        tab2.run.set('40123 40127')
        tab2.get_data()
        
        try:
            function(*args, **kwargs, tab=tab, b=b, tab2=tab2)
        finally:
            b.on_closing()
            del b
            
    return wrapper

@with_bfit
def test_populate(b=None, tab=None, tab2=None):

    # check populate
    tab.populate()
    assert('2020.40123' in tab.fit_lines.keys()), 'First fetched run populated'
    assert('2020.40127' in tab.fit_lines.keys()), 'Second fetched run populated'
    
    # remove
    tab2.data_lines['2020.40123'].degrid()
    tab.populate()
    assert('2020.40123' not in tab.fit_lines.keys()), 'First fetched run removed'
    
    # fetch again
    tab2.get_data()
    tab.populate()
    assert('2020.40123' in tab.fit_lines.keys()), 'Re-populated after run removal and refetch'

@with_bfit    
def test_populate_param(b=None, tab=None, tab2=None):
        
    tab.populate()
        
    # get fit line
    line = tab.fit_lines['2020.40123']
    
    # set fit fn
    tab.fit_function_title.set('Str Exp')
    tab.populate_param(force_modify=True)
    assert_equal(tuple(line.data.fitpar.index), ('1_T1', 'amp', 'beta'), 'Single function parameters populated')
    
    # multiple terms
    tab.fit_function_title.set('Exp')
    tab.n_component.set(2)
    tab.populate_param(force_modify=True)
    assert_equal(tuple(line.data.fitpar.index), ('1_T1_0', '1_T1_1', 'amp_0', 'amp_1'), 
                 'Two term function parameters populated')
                 
    # undo changes
    tab.n_component.set(1)
    
# do the fit separately
def test_fit_curve_fit():       fit(separate_curve_fit, 'curve_fit')
def test_fit_migrad():          fit(separate_migrad, 'migrad_hesse')
def test_fit_minos():           fit(separate_minos, 'migrad_minos')
def test_fit_single_curve_fit():fit_single(separate_curve_fit, 'curve_fit')
def test_fit_single_migrad():   fit_single(separate_migrad, 'migrad_hesse')
def test_fit_single_minos():    fit_single(separate_minos, 'migrad_minos')

def separate_curve_fit(entry, run, **kwargs):
    
    # data
    year, run = tuple(map(int, run.split('.')))
    dat = bd.bdata(run, year)
    
    # fit function
    pexp = pulsed_exp(lifetime = bd.life.Li8, pulse_len = dat.pulse_s)

    p0 = [float(entry[k]['p0'].get()) for k in entry.keys()]
    blo = [float(entry[k]['blo'].get()) for k in entry.keys()]
    bhi = [float(entry[k]['bhi'].get()) for k in entry.keys()]
    
    # fit
    t,a,da = dat.asym('c')
    par, cov = curve_fit(pexp, t, a, sigma=da, absolute_sigma=True, p0=p0, 
                         bounds=[blo, bhi], **kwargs)
    std = np.diag(cov)**0.5
    
    # chi2
    chi = np.sum(((a-pexp(t, *par))/da)**2)/(len(a)-2)
    
    return (par, std, std, chi)

def separate_migrad(entry, run, do_minos=False, **kwargs):
    
    # data
    year, run = tuple(map(int, run.split('.')))
    dat = bd.bdata(run, year)
    
    # fit function
    pexp = pulsed_exp(lifetime = bd.life.Li8, pulse_len = dat.pulse_s)

    # get p0, bounds
    p0 = {k.replace('1_T1','lambda_s'):float(entry[k]['p0'].get()) for k in entry.keys()}
    bounds = [( float(entry[k]['blo'].get()), 
                float(entry[k]['bhi'].get())) for k in entry.keys()]
    
    # fit
    t,a,da = dat.asym('c')
    m = minuit(pexp, t, a, da, **p0, limit = bounds, **kwargs)
    m.migrad()
    
    if do_minos: m.minos()
    
    par = m.values
    
    if do_minos: 
        n = len(par)
        lower = np.abs(np.array([m.merrors[i].lower for i in range(n)]))
        upper = np.array([m.merrors[i].upper for i in range(n)])
    else:
        lower = m.errors
        upper = lower
        
    chi = m.chi2
        
    return (par, lower, upper, chi)

def separate_minos(entry, run, **kwargs):
    return separate_migrad(entry, run, do_minos=True, **kwargs)

@with_bfit    
def fit(separate_fit, minimizer, b=None, tab=None, tab2=None):
    
    # get data
    tab.populate()
    
    # set minimizer
    b.minimizer.set('bfit.fitting.fitter_%s' % minimizer)
    b.set_fit_routine()
    
    # do fit in bfit
    tab.do_fit()
    
    # check return values -----------------------------------------------------
    
    for k, line in tab.fit_lines.items():
    
        entry = {l.pname: l.variable for l in line.lines}
        
        # get results separately
        par, lower, upper, chi2 = separate_fit(entry, k)    
        
        # get results displayed
        res = [float(entry[k]['res'].get()) for k in entry.keys()]
        low = [float(entry[k]['dres-'].get()) for k in entry.keys()]
        upp = [float(entry[k]['dres+'].get()) for k in entry.keys()]
        
        # check
        assert_array_almost_equal(par, res, err_msg = 'Copying %s minimizer results for %s' % (minimizer, k), decimal=3)
        assert_array_almost_equal(lower, low, err_msg = 'Copying %s minimizer lower errors for %s' % (minimizer, k), decimal=3)
        assert_array_almost_equal(upper, upp, err_msg = 'Copying %s minimizer upper errors for %s' % (minimizer, k), decimal=3)
        
        # check chi2
        chi = float(entry['1_T1']['chi'].get())
        assert_almost_equal(chi, chi2, 
                        err_msg = 'Copying chi2 for %s minimizer for %s' % (minimizer, k), 
                        decimal=2)
    
    # reset minimizer
    b.minimizer.set('bfit.fitting.fitter_curve_fit')
    b.set_fit_routine()

@with_bfit
def fit_single(separate_fit, minimizer, b=None, tab=None, tab2=None):
    
    # clear setup
    tab2.remove_all()
    
    # fetch
    tab2.year.set(2020)
    tab2.run.set('40123')
    tab2.get_data()

    tab.populate()
    line = tab.fit_lines['2020.40123']
    entry = {l.pname: l.variable for l in line.lines}
    
    # set minimizer
    b.minimizer.set('bfit.fitting.fitter_%s' % minimizer)
    b.set_fit_routine()
    
    # check return value ------------------------------------------------------
    
    # get results separately
    par, lower, upper, chi2 = separate_fit(entry, '2020.40123')    
    tab.do_fit()
    
    # get results displayed
    res = [float(entry[k]['res'].get()) for k in entry.keys()]
    low = [float(entry[k]['dres-'].get()) for k in entry.keys()]
    upp = [float(entry[k]['dres+'].get()) for k in entry.keys()]
    
    # check
    assert_array_almost_equal(par, res, 
                err_msg = 'Copying %s minimizer results for single run' % minimizer, decimal=3)
    assert_array_almost_equal(lower, low, 
                err_msg = 'Copying %s minimizer lower errors for single run' % minimizer, decimal=3)
    assert_array_almost_equal(upper, upp, 
                err_msg = 'Copying %s minimizer upper errors for single run' % minimizer, decimal=3)
    
    # check chi2
    chi = float(entry['1_T1']['chi'].get())
    assert_almost_equal(chi, chi2, err_msg='Copying chi2 for %s minimizer for single run' % minimizer, decimal=2)
    
    # reset minimizer
    b.minimizer.set('bfit.fitting.fitter_curve_fit')
    b.set_fit_routine()

@with_bfit
def test_fit_fn_name(b=None, tab=None, tab2=None):
    
    # set rebin
    tab2.data_lines['2020.40123'].degrid()
    
    # set function
    tab.fit_function_title.set('Bi Exp')
    
    # fit
    tab.populate()    
    tab.do_fit()
    
    # fit input: fn_name, ncomp, data_list
    # data_list: [bdfit, pdict, doptions]
    
    # check function
    assert_equal(tab.fit_input[0], 'Bi Exp', "Function name passing to fitter")

@with_bfit
def test_fit_rebin(b=None, tab=None, tab2=None):
    
    # set rebin
    tab2.data_lines['2020.40123'].rebin.set(10)
    tab2.data_lines['2020.40127'].rebin.set(20)
    
    # fit
    tab.populate()    
    tab.use_rebin.set(True)
    tab.do_fit()
    
    # fit input: fn_name, ncomp, data_list
    # data_list: [bdfit, pdict, doptions]
    
    # check
    assert_equal(tab.fit_input[2][0][2]['rebin'], 10, "Rebin passing for file 1")
    assert_equal(tab.fit_input[2][1][2]['rebin'], 20, "Rebin passing for file 2")

@with_bfit
def test_fit_ncomp(b=None, tab=None, tab2=None):
    
    # set rebin
    tab2.data_lines['2020.40123'].rebin.set(10)
    tab2.data_lines['2020.40127'].rebin.set(20)
    tab.populate()    
    
    # set ncomp
    tab.n_component.set(2)
    tab.populate_param(force_modify=True)    
    
    # fit
    tab.populate()    
    tab.do_fit()
    
    # fit input: fn_name, ncomp, data_list
    # data_list: [bdfit, pdict, doptions]
    
    # check ncomp
    assert_equal(tab.fit_input[1], 2, "Number of components passing to fitter")

@with_bfit
def test_shared(b=None, tab=None, tab2=None):
    
    # get data
    tab.populate()
    line = tab.fit_lines['2020.40123']
    line2 = tab.fit_lines['2020.40127']
    entry = {l.pname: l.variable for l in line.lines}
    entry2 = {l.pname: l.variable for l in line2.lines}
    
    # shared input
    entry['1_T1']['shared'].set(True)
    tab.do_fit()
    
    T11 = float(entry['1_T1']['res'].get())
    T12 = float(entry2['1_T1']['res'].get())
    
    assert_almost_equal(T11, T12, err_msg = 'Shared result not equal')
    
    T11 = float(entry['1_T1']['dres-'].get())
    T12 = float(entry2['1_T1']['dres-'].get())
    
    assert_almost_equal(T11, T12, err_msg = 'Shared lower error not equal')
    
    T11 = float(entry['1_T1']['dres+'].get())
    T12 = float(entry2['1_T1']['dres+'].get())
    
    assert_almost_equal(T11, T12, err_msg = 'Shared upper error not equal')
    
    # check the unshared result
    amp1 = float(entry['amp']['res'].get())
    amp2 = float(entry2['amp']['res'].get())
    
    assert amp1 != amp2, 'Unshared result is equal'
    
    amp1 = float(entry['amp']['dres-'].get())
    amp2 = float(entry2['amp']['dres-'].get())
    
    assert amp1 != amp2, 'Unshared lower error not equal'
    
    amp1 = float(entry['amp']['dres+'].get())
    amp2 = float(entry2['amp']['dres+'].get())
    
    assert amp1 != amp2, 'Unshared upper error not equal'
    
    # unshare
    entry['1_T1']['shared'].set(False)
    
@with_bfit
def test_modify_for_all_reset_p0(b=None, tab=None, tab2=None):
    
    # get data
    tab.populate()
    line = tab.fit_lines['2020.40123']
    line2 = tab.fit_lines['2020.40127']
    entry = {l.pname: l.variable for l in line.lines}
    entry2 = {l.pname: l.variable for l in line2.lines}
    
    # modify
    tab.set_as_group.set(True)
    
    initial = {}
    for k in entry.keys():    
        initial[k] = {}
        for c in ('p0', 'blo', 'bhi'):
            
            # get initial state
            initial[k][c] = float(entry[k][c].get())
            
            # modify all
            entry[k][c].set('25')
            assert_equal(entry2[k][c].get(), '25', err_msg = "Modify all for %s (%s)" % (k, '25'))
            
    # reset
    tab.set_as_group.set(False)
    tab.do_reset_initial()
    for k in entry.keys():    
        for c in ('p0', 'blo', 'bhi'):
            assert_equal(float(entry[k][c].get()), initial[k][c], 
                        err_msg = "Reset p0 for %s (%s)" % (k, c))
                
@with_bfit
def test_fixed(b=None, tab=None, tab2=None):
    
    # get data
    tab.populate()
    line = tab.fit_lines['2020.40123']
    entry = {l.pname: l.variable for l in line.lines}
    
    # fixed input -----------------------------------------------------------
    entry['1_T1']['p0'].set('1')
    entry['1_T1']['fixed'].set(True)
    tab.do_fit()
    
    assert_equal(float(entry['1_T1']['res'].get()), 1, 'Fixed result')
    assert entry['1_T1']['dres-'].get() == '', 'Fixed lower error'
    assert entry['1_T1']['dres+'].get() == '', 'Fixed upper error'
    assert float(entry['amp']['res'].get()) != 1, 'Unfixed result'
    assert not np.isnan(float(entry['amp']['dres-'].get())), 'Unfixed lower error'
    assert not np.isnan(float(entry['amp']['dres+'].get())), 'Unfixed upper error'
    
    # unfix
    entry['1_T1']['fixed'].set(False)

@with_bfit
def test_p0_prior(b=None, tab=None, tab2=None):
    
    # get data
    tab.populate()
    
    # fit
    tab.do_fit()
    b.do_close_all()
    
    # set next p0
    tab.set_prior_p0.set(True)
    
    # get more data
    tab2.run.set('40124')
    tab2.get_data()
    tab.populate()
    
    # get results and compare
    result = tab.fit_lines['2020.40127'].lines
    p0 = tab.fit_lines['2020.40124'].lines
    
    for r, p in zip(result, p0):
        assert_almost_equal(float(r.get('res')), float(p.get('p0')), 
             err_msg = 'Set last result as p0 for new run for %s'%r.pname, decimal=5)
    
@with_bfit
def test_result_as_p0(b=None, tab=None, tab2=None):
    
    # get data
    tab.populate()
    
    # fit
    tab.do_fit()
    b.do_close_all()
    
    # set result as P0
    tab.do_set_result_as_initial()
    
    # check
    for run_id in tab.fit_lines.keys():
        entry = {l.pname: l.variable for l in tab.fit_lines[run_id].lines}
    
        for k in entry.keys():
            assert_almost_equal(float(entry[k]['res'].get()), float(entry[k]['p0'].get()), 
                err_msg = 'Set result as p0 for %s'%k, decimal=5)

def check_sharing_assignment(tab, fline, test_name):
    
    # check all shared param independent
    for k1, v1 in tab.share_var.items():
        for k2, v2 in tab.share_var.items():
            if k1 != k2: 
                assert v1 != v2, \
                    'share_var BooleanVar equal for '+\
                    '"{}" and "{}" during {}'.format(k1, k2, test_name)
                
    # check all lines param assigned correctly
    for l1 in fline.lines:
        for l2 in fline.lines:
            if l1 != l2:
                assert l1.variable['shared'] != l2.variable['shared'], \
                    'line variable[shared] BooleanVar equal for '+\
                    '"{}" and "{}" during {}'.format(k1, k2, test_name)
    
@with_bfit
def test_sharing_assignment(b=None, tab=None, tab2=None):
    """Handling of sharing paramters not being independent"""
    
    tab.populate()
    fline = tab.fit_lines['2020.40123']
    
    # check biexp
    tab.fit_function_title.set('Bi Exp')
    tab.populate_param(force_modify=True)
    check_sharing_assignment(tab, fline, 'biexp fn loaded')
    
    # check three exp
    tab.fit_function_title.set('Exp')
    tab.n_component.set(3)
    tab.populate_param(force_modify=True)
    check_sharing_assignment(tab, fline, 'three exp fn loaded')
        
@pytest.mark.filterwarnings('ignore:Tight layout')
@pytest.mark.filterwarnings('ignore:Warning')
@with_bfit
def test_draw_fit_results(b=None, tab=None, tab2=None):
    
    # get data
    tab.populate()
    
    # fit
    tab.do_fit()
    b.do_close_all()
    
    # get list of draw-able parameters
    values = tab.xaxis_combobox['values'][1:]
    
    # draw 
    tab.xaxis.set(values[0])
    for v in values[1:]:
        tab.yaxis.set(v)
        tab.draw_param()
        b.plt.clf('param')
    
    # annotation
    tab.annotation.set(values[1])
    tab.draw_param()
    b.do_close_all()
    
    # if nothing failed then we're ok!
    




from tkinter import *
from tkinter import ttk, messagebox
from functools import partial
import logging
import iminuit.pdg_format as pdg

import numpy as np

from bfit import logger_name
import bfit.backend.colors as colors

class InputLine(object):
    """
        Stores one line of inputs:
            'p0', 'blo', 'bhi', 'res', 'dres-', 'dres+', 'chi', 'fixed', 'shared'
        as well as the frames and variables

        bfit: bfit object
        data: fitdata object
        entry: dict[col] = entry or checkbutton object
        frame: ttk.Frame
        label: ttk.label, parameter name
        logger: logger
        pname: string, name of parameter for this line
        variable: dict[col] = variable (StringVar or BooleanVar)
    """

    width = 13
    width_chi = 7
    columns = ['p0', 'blo', 'bhi', 'res', 'dres-', 'dres+', 'chi', 'fixed', 'shared']

    # ======================================================================= #
    def __init__(self, frame, bfit, data, show_chi=False):
        """
            frame: frame in which line will grid (first row is 1)
            show_chi: if true, grid the chi parameter
        """

        # get logger
        self.logger = logging.getLogger(logger_name)
        self.logger.debug('Initializing')

        # assign inputs and defaults
        self.pname = ''
        self.bfit = bfit
        self.data = data

        self.frame = frame
        self.label = ttk.Label(self.frame, text=self.pname, justify='right')

        self.variable = {   'p0': StringVar(),
                            'blo': StringVar(),
                            'bhi': StringVar(),
                            'res': StringVar(),
                            'dres-': StringVar(),
                            'dres+': StringVar(),
                            'chi': StringVar(),
                            'fixed': BooleanVar(),
                            'shared': BooleanVar(),
                        }

        self.entry = {}
        for key, var in self.variable.items():

            # stringvar
            if key not in ('fixed', 'shared'):

                if key == 'chi':
                    width = self.width_chi
                else:
                    width = self.width

                self.entry[key] = Entry(self.frame,
                                          textvariable=self.variable[key],
                                          width=width)

            # booleanvar
            else:
                self.entry[key] = ttk.Checkbutton(self.frame,
                                                  text='',
                                                  variable=self.variable[key],
                                                  onvalue=True,
                                                  offvalue=False)

            # set colors and state
            if key in ('res', 'dres-', 'dres+', 'chi'):
                self.entry[key]['state'] = 'readonly'
                self.entry[key]['foreground'] = colors.foreground

        # disallow fixed and shared variables
        self.variable['shared'] = self._set_trace(self.variable['shared'],
                                                  'unfix', self._unfix)
        self.variable['fixed'] = self._set_trace(self.variable['fixed'],
                                                 'unshare', self._unshare)

        # set modify all synchronization
        for k in ('p0', 'blo', 'bhi', 'fixed'):

            # set new trace callback
            self.variable[k] = self._set_trace(self.variable[k],
                                               'modify_all',
                                               partial(self._modify_all, col=k))

    # ======================================================================= #
    def _modify_all(self, *_, col):
        """
            Do modify all synchronization. Make other lines of the same id equal
            in value
        """

        # check if enabled
        if not (self.bfit.fit_files.set_as_group.get() or self.variable['shared'].get()):
            return

        # disable trace
        self.variable[col], tr = self._pop_trace(self.variable[col], 'modify_all')

        # set
        self.bfit.fit_files.set_lines(pname=self.pname,
                                      col=col,
                                      value=self.variable[col].get(),
                                      skipline=self)

        # re-enable trace
        self.variable[col] = self._set_trace(self.variable[col], 'modify_all', tr)

    # ======================================================================= #
    def _pop_trace(self, var, name):
        """
            Remove and return trace function
        """

        # check if variable has trace dict
        if not hasattr(var, 'trace_id') or name not in var.trace_id.keys():
            return (var, None)

        # pop
        tr = var.trace_id[name]
        var.trace_remove("write", tr[0])
        del var.trace_id[name]

        # return
        return (var, tr[1])

    # ======================================================================= #
    def _set_trace(self, var, name, function):
        """
            Set the trace functions
        """

        # no input
        if function is None:
            return var

        # check if variable has trace dict
        if not hasattr(var, 'trace_id'):
            var.trace_id = {}

        # check if trace exists
        self._pop_trace(var, name)

        # add trace
        var.trace_id[name] = (var.trace_add("write", function), function)

        return var

    # ======================================================================= #
    def _unfix(self, *_):
        """
            disallow fixed shared parameters
        """
        if self.variable['shared'].get():
            self.variable['fixed'].set(False)

    # ======================================================================= #
    def _unshare(self, *_):
        """
            disallow fixed shared parameters
        """
        if self.variable['fixed'].get():
            self.variable['shared'].set(False)

    # ======================================================================= #
    def assign_inputs(self):
        """
            Make sure the inputs are saved to fitdata.fitpar DataFrame
        """

        # check if valid
        if not self.pname:
            return

        # assign
        gen = self.data.gen_set_from_var
        for k in ('p0', 'blo', 'bhi', 'fixed', 'shared'):

            # set new trace callback
            self.variable[k] = self._set_trace(self.variable[k],
                                               'sync_fitpar',
                                               gen(self.pname, k, self.variable[k]))

    # ======================================================================= #
    def assign_shared(self):
        """
            Link the shared values
        """

        # no key
        if not self.pname:
            return

        # get dict of shared boolean var
        share_var = self.bfit.fit_files.share_var

        # check if key is present
        if self.pname not in share_var.keys():
            share_var[self.pname] = BooleanVar()

        # assign key
        self.variable['shared'] = share_var[self.pname]

        # set trace to uncheck fixed box
        self.variable['shared'] = self._set_trace(self.variable['shared'],
                                                  'unfix_{id}'.format(id=self.data.id),
                                                  self._unfix)

        # link to checkbox
        self.entry['shared'].config(variable=self.variable['shared'])

    # ======================================================================= #
    def degrid(self):
        """
            Remove the entries
        """
        self.label.destroy()

        for i, key in enumerate(self.columns):
            self.entry[key].destroy()

    # ======================================================================= #
    def disable(self):
        """
            Prevent editing
        """
        for k, e in self.entry.items():
            if k not in ('chi',):
                e.configure(state='disabled')

    # ======================================================================= #
    def enable(self):
        """
            Allow editing
        """
        for k, e in self.entry.items():
            if k in ('res', 'dres+', 'dres-', 'chi'):
                e.configure(state='readonly')
            else:
                e.configure(state='normal')

    # ======================================================================= #
    def get(self, col):
        """
            get values

            col: str, name of column to get
        """

        # wildcards
        if col == '*':
            return {c:self.get(c) for c in self.columns}

        # get single value
        if col in self.variable.keys():
            v = self.variable[col].get()

        if type(v) is str:

            if v == '':
                return np.nan

            try:
                v = float(v)
            except ValueError as errmsg:
                self.logger.exception("Bad input.")
                messagebox.showerror("Error", str(errmsg))
                raise errmsg

        return v

    # ======================================================================= #
    def grid(self, row):
        """
            Grid the entries
        """
        self.label.grid(column=0, row=row, sticky='e')

        for i, key in enumerate(self.columns):

            if not row == 2 and key == 'chi':
                pass
            elif key == 'chi':
                self.entry[key].grid(column=i+1, row=row, rowspan=100, padx=5)
            else:
                self.entry[key].grid(column=i+1, row=row, padx=5)

    # ======================================================================= #
    def set(self, pname=None, **values):
        """
            set values

            pname: string, parameter name (ex: 1_T1)
            values: keyed by self.columns, the numerical or boolean values for each
                    column to take
        """

        # label
        if pname is not None:
            self.pname = pname
            self.label.config(text=pname)

            # assign inputs
            self.assign_shared()
            self.assign_inputs()

            # check if line is constrained
            constr_set = self.bfit.fit_files.pop_fitconstr.constraints_are_set
            if pname in self.data.constrained.keys() and constr_set:
                self.disable()
                self.variable['fixed'].set(False)
                self.variable['shared'].set(False)
            else:
                self.enable()

        # set values
        for k, v in values.items():

            # disable traces
            self.variable[k], tr = self._pop_trace(self.variable[k], 'modify_all')

            # blank
            if str(v) == 'nan':

                # set as blank
                if type(self.variable[k]) == StringVar:
                    self.variable[k].set('')

                # set chi box color
                if k == 'chi':
                    self.entry['chi']['readonlybackground']=colors.readonly

            # set boolean
            elif type(v) is bool:
                self.variable[k].set(v)

            # set string
            elif type(v) is str:
                self.variable[k].set(v)

            # set float
            else:

                v = float(v)

                # chisq
                if k == 'chi':

                    # set number decimal places
                    n_figs = 2

                    # set color
                    if v > self.bfit.fit_files.chi_threshold:
                        self.entry['chi']['readonlybackground']='red'
                    else:
                        self.entry['chi']['readonlybackground']=colors.readonly

                # results and errors rounding
                elif k == 'res':
                    string = pdg.pdg_format(v, values['dres+'], values['dres-'],
                                            format=(' %s',   # str: format spec for lower asymmetric error
                                                    ' %s',   # str: format spec for upper asymmetric error
                                                    ' %s',   # str: format spec for symmetric error
                                                    '%s',    # str: format spec for label
                                                    '%se%i',# str: format spec for scientific notation
                                                    True,    # bool: whether to strip trailing zeros and dots
                                                    None)    # tuple of str OR None: replacement for 'nan' and 'inf
                                            )
                    # get values to set from string
                    res, dresp, dresm = string.split(' ')
                    if 'e' in dresm:
                        dresm, sci = dresm.split('e')
                        res = f'{res}e{sci}'
                        dresp = f'{dresp}e{sci}'
                        dresm = f'{dresm}e{sci}'
                    self.variable['res'].set(res)
                    self.variable['dres+'].set(dresp)
                    self.variable['dres-'].set(dresm)


                elif k in ('dres+', 'dres-'):
                    continue

                # other value
                else:
                    n_figs = self.bfit.rounding

                # round to n_figs significant figures decimal places
                if k not in ('res', 'dres+', 'dres-'):
                    try:
                        v_decimal = v - int(v)
                        v_decimal = float('{:.{p}g}'.format(v_decimal, p=n_figs))
                        v = int(v) + v_decimal
                    except OverflowError:
                        pass
                    self.variable[k].set('{:.{p}g}'.format(v, p=8))

            # set traces
            self.variable[k] = self._set_trace(self.variable[k], 'modify_all', tr)
            self.variable[k] = self._set_trace(self.variable[k], 'sync_fitpar',
                                                self.data.gen_set_from_var(self.pname,
                                                                           k,
                                                                           self.variable[k]))



# test inspect tab
# Derek Fujimoto
# Feb 2021

from numpy.testing import *
import numpy as np
from bfit.gui.popup_units import popup_units
from bfit.gui.bfit import bfit

# filter unneeded warnings
import pytest
pytestmark = pytest.mark.filterwarnings('ignore:2020')

def with_bfit(year, run):
    def wrapper2(function):
        def wrapper(*args, **kwargs):
            # make gui
            b = bfit(None, True)
            tab = b.fileviewer
            
            tab.year.set(year)
            tab.runn.set(run)
            tab.get_data()
            
            units = popup_units(b)
            
            try:
                return function(*args, **kwargs, b=b, tab=tab, units=units)
            finally:
                b.on_closing()
                del b      
        return wrapper
    return wrapper2
    
@with_bfit(year=2020, run=40010)
def test_units_rescale(b=None, tab=None, units=None):
    units.input['1n'][0].set('1')
    units.set()
    tab.draw('inspect')
    ax = b.plt.gca('inspect')    
    assert ax.get_xticks()[0] == 40000, "Data rescaled by 1000"

@with_bfit(year=2020, run=40010)
def test_units_rename(b=None, tab=None, units=None):
    test_unit = 'test_unit'
    
    units.input['1n'][1].set('test_unit')
    units.set()
    tab.draw('inspect')
    ax = b.plt.gca('inspect')    
    assert ax.get_xlabel()[-len(test_unit)-1:-1] == test_unit, \
            "Units rename %s != %s" % (ax.get_xlabel(), test_unit)

@with_bfit(year=2020, run=40010)
def test_units_defaults_legacy(b=None, tab=None, units=None):
    units.input['1n'][1].set('default')
    units.set()
    tab.draw('inspect')
    ax = b.plt.gca('inspect')    
    assert ax.get_xlabel() == 'Voltage (default)', "xlabel not Voltage (default), is: %s" % ax.get_xlabel()
    assert ax.get_xticks()[0] == 40000, "Data not scaled to mV, tick zero is: %s" % str(ax.get_xticks()[0])

# umcomment when a run is saved which this can be run on. 
# ~ @with_bfit(year=2020, run=40010)
# ~ def test_units_defaults_new(b=None, tab=None, units=None):
    # ~ units.input['1n'][1].set('default')
    # ~ units.set()
    # ~ tab.draw('inspect')
    # ~ ax = b.plt.gca('inspect')    
    # ~ assert ax.get_xlabel() == 'Voltage (V)', "xlabel not Voltage (V), is: %s" % ax.get_xlabel()
    # ~ assert ax.get_xticks()[0] == 40, "Data not scaled to volts, tick zero is: %s" % str(ax.get_xticks()[0])    



# Set function paramters from gui window
# Derek Fujimoto
# April 2019

from tkinter import *
from tkinter import ttk, messagebox
from bfit import logger_name
from bfit.backend.FunctionPlacer import FunctionPlacer
from bfit.fitting.decay_31mg import fa_31Mg
import bfit.fitting.functions as fns

import matplotlib.pyplot as plt
import logging
import bdata as bd
import numpy as np

# ========================================================================== #
class popup_param(object):
    """
        Popup window for graphically finding input parameters.

        bhi:            dictionary of StringVar objects to link parameters
        blo:            dictionary of StringVar objects to link parameters
        data:           bdata object
        fig:            maplotlib figure object
        fitter:         fit_tab.fitter obje (defined in default_routines.py)
        first:          if True, first time through fixing parameters
        fname:          name of the function
        lines:          dict of InputLine objects to link parameters
        logger:         logging variable
        mode:           1 or 2 to switch between run modes
        n_components:   number of components in the fit function
        parnames:       tuple of names of the parameters
        p0:             dictionary of StringVar objects to link parameters
        selection:      StringVar, track run selection
        win:            TopLevel window
        xy:             (x, asym, dasym) tuple
    """

    # parameter mapping
    parmap = {  '1_T1':'lam',
                '1_T1b':'lamb',
                'amp':'amp',
                'beta':'beta',
                'fraction_b':'fraction_b',
                'baseline':'base',
                'peak':'peak',
                'fwhmA':'fwhm',
                'fwhmB':'fwhm',
                'fwhmL':'fwhm',
                'height':'amp',
                'heightA':'amp',
                'heightB':'amp',
                'sigma':'fwhm',
                'fracL':'fracL',
                'mean':'peak',
                'amp0':'amp0',
                'amp1':'amp1',
                'amp2':'amp2',
                'amp3':'amp3',
                'nu_0':'nu_0',
                'nu_q':'nu_q',
                'efgAsym':'eta',
                'efgPhi':'phi',
                'efgTheta':'theta',
                'fwhm':'fwhm',
             }

    # ====================================================================== #
    def __init__(self, bfit, id=''):
        self.bfit = bfit

        # get logger
        self.logger = logging.getLogger(logger_name)
        self.logger.info('Initializing gui param popup')

        # make a new window
        self.win = Toplevel(bfit.mainframe)
        self.win.title('Find P0')
        frame = ttk.Frame(self.win, relief='sunken', pad=5)

        # icon
        self.bfit.set_icon(self.win)

        # Labels
        ttk.Label(frame, text="Select Run").grid(column=0, row=0, sticky=E)

        # box for run select
        self.selection = StringVar()
        select_box = ttk.Combobox(frame, textvariable=self.selection,
                                  state='readonly')
        select_box.bind('<<ComboboxSelected>>', self.setup)

        # get run list
        runlist = list(self.bfit.fit_files.fit_lines.keys())
        runlist.sort()
        # ~ runlist = ['('+r.split('.')[0]+') '+r.split('.')[1] for r in runlist]
        select_box['values'] = runlist

        # gridding
        frame.grid(column=0, row=1, sticky=(N, W, E, S))
        select_box.grid(column=0, row=1, sticky=E)

        # start looking automatically
        if id:
            self.selection.set(id)
            self.setup()
            self.win.withdraw()

    # ====================================================================== #
    def setup(self, *args):
        """Get parameters for placing function and start the run squence"""

        # get run selection
        run_id = self.selection.get()
        self.logger.info('Running P0 GUI finder on run %s', run_id)

        # get data
        self.data = self.bfit.data[run_id]
        mode = self.data.mode

        # mode switching
        if mode in ('20', '2h'):         self.mode = 2
        elif mode in ('1f', '2e', '1w', '1x'):  self.mode = 1
        else:
            self.logger.warning('P0 Finder not configured for run mode %s', mode)
            print('P0 Finder not configured for run mode %s'%mode)

        # make new window
        self.fig = plt.figure()
        ax = self.fig.add_subplot(111)

        # draw data
        omit = self.data.omit.get()
        if omit == self.bfit.fetch_files.bin_remove_starter_line:
            omit = ''
        self.xy = self.data.asym(self.bfit.get_asym_mode(self.bfit.fit_files),
                                 rebin=self.data.rebin.get(), omit=omit)
        ax.errorbar(*self.xy, fmt='.', color='k', ecolor='k')

        # plot elements - don't do tight_layout here - blocks matplotlib signals
        ax.set_ylabel('Asymmetry')
        if self.mode == 2:     ax.set_xlabel('Time')
        elif self.mode == 1:   ax.set_xlabel('Frequency')
        self.fig.show()

        # get parameters list and run finder
        fit_tab = self.bfit.fit_files
        self.fitter = fit_tab.fitter
        self.n_components = fit_tab.n_component.get()
        self.fname = fit_tab.fit_function_title.get()
        self.parnames = self.fitter.gen_param_names(fn_name=self.fname,
                                          ncomp=self.n_components)

        lines = fit_tab.fit_lines[run_id].lines
        self.lines = {line.pname:line  for line in lines}

        # make initial parameter list
        self.p0 =  {k:v.variable['p0']  for k, v in self.lines.items()}
        self.blo = {k:v.variable['blo'] for k, v in self.lines.items()}
        self.bhi = {k:v.variable['bhi'] for k, v in self.lines.items()}

        # check for zero width
        for k in self.p0.keys():
            if 'fwhm' in k or 'sigma' in k:
                if float(self.p0[k].get()) == 0:
                     self.p0[k].set('1e-9')
        self.run()

    # ====================================================================== #
    def split_components(self, mixed):
        """
            Split a dictionary of parameter components into lists of the base
            names to be used in invididual functions
        """

        split = []
        if self.n_components > 1:
            for i in range(self.n_components):
                element = {}
                for k in mixed.keys():
                    if 'base' in k:
                        element[self.parmap[k]] = mixed[k]
                    elif str(i) in k:
                        element[self.parmap['_'.join(k.split('_')[:-1])]] = mixed[k]
                split.append(element)
        else:
            split.append({self.parmap[k]:mixed[k] for k in mixed.keys()})

        return split

    # ====================================================================== #
    def run(self):
        """
            Run the function placer

            comp = component number to run
        """

        try:
            del self.fplace
        except AttributeError:
            pass
        else:
            self.fig.axes[0].cla()
            self.fig.axes[0].errorbar(*self.xy, fmt='.')

        # ensure matplotlib signals work. Not sure why this is needed.
        self.fig.tight_layout()

        # get paramters, translating the names
        p0 =  self.split_components(self.p0)
        blo = self.split_components(self.blo)
        bhi = self.split_components(self.bhi)

        # get fitting function
        if self.fname == 'Lorentzian':
            fn = fns.lorentzian     # freq, peak, fwhm, amp
        elif self.fname == 'BiLorentzian':
            fn = lambda freq, peak, fwhm, amp : fns.bilorentzian(freq, peak, fwhm, amp/3, fwhm/3, amp*2/3)
        elif self.fname == 'Gaussian':
            fn = lambda freq, peak, fwhm, amp : fns.gaussian(freq, peak, fwhm, amp)
        elif self.fname == 'QuadLorentz':
            fn = lambda freq, nu_0, nu_q, eta, theta, phi, \
                        amp0, amp1, amp2, amp3, fwhm: \
                        fns.quadlorentzian(freq, nu_0, nu_q, eta, theta, phi, \
                        amp0, amp1, amp2, amp3, \
                        fwhm, fwhm, fwhm, fwhm,
                        I = self.fitter.spin[self.fitter.probe_species])
        elif self.fname == 'PseudoVoigt':
            fn = lambda freq, peak, fwhm, amp, fracL : fns.pseudo_voigt(freq, peak, fwhm, amp, fracL)

        elif self.fname in ('Exp', 'Str Exp'):

            # get function
            pulse = self.data.pulse_s
            lifetime = bd.life[self.bfit.probe_species.get()]

            if self.fname == 'Exp':
                f1 = fns.pulsed_exp(lifetime=lifetime, pulse_len=pulse)

                if self.bfit.probe_species.get() == 'Mg31':
                    fn = lambda x, lam, amp : fa_31Mg(x, pulse)*f1(x, lam, amp)
                else:
                    fn = lambda x, lam, amp : f1(x, lam, amp)

            elif self.fname == 'Str Exp':
                f1 = fns.pulsed_strexp(lifetime=lifetime, pulse_len=pulse)

                if self.bfit.probe_species.get() == 'Mg31':
                    fn = lambda x, lam, amp, beta : fa_31Mg(x, pulse)*f1(x, lam, beta, amp)
                else:
                    fn = lambda x, lam, amp, beta : f1(x, lam, beta, amp)
        else:
            self.cancel()
            errormsg = 'Function "%s" not implemented in P0 Finder' % self.fname
            self.logger.warning(errormsg)
            messagebox.showerror("Error", errormsg)
            raise RuntimeError(errormsg)

        self.fig.canvas.mpl_connect('close_event', self.cancel)

        self.fplace = FunctionPlacer(fig=self.fig,
                                     data=self.data,
                                     fn_single=fn,
                                     ncomp=self.n_components,
                                     p0=p0,
                                     fnname=self.fname,
                                     asym_mode=self.bfit.get_asym_mode(self.bfit.fit_files),
                                     endfn=self.endfn,
                                     spin=self.fitter.spin[self.fitter.probe_species])

    # ====================================================================== #
    def endfn(self, p0, base):
        """Set output fields"""

        for k in self.p0.keys():

            # skip baseline
            if 'base' in k:
                continue

            # single component
            if len(p0) == 1:
                i = 0
                key = k

            # multi component
            else:
                s = k.split('_')
                key = '_'.join(s[:-1])
                i = int(s[-1])

            # set p0
            val = p0[i][self.parmap[key]]
            self.lines[k].set(p0=val)

            # check bounds are ok
            blo = float(self.blo[k].get())
            bhi = float(self.bhi[k].get())

            if val < blo:   self.blo[k].set('-inf')
            elif val < bhi: self.bhi[k].set('inf')

        # baseline
        if 'baseline' in self.p0.keys():
            self.p0['baseline'].set(base)

            # check bounds are ok
            blo = float(self.blo['baseline'].get())
            bhi = float(self.bhi['baseline'].get())

            if base < blo:   self.blo['baseline'].set('-inf')
            elif base < bhi: self.bhi['baseline'].set('inf')

    # ====================================================================== #
    def cancel(self, *args):
        if hasattr(self, 'fplace'):  del self.fplace
        if hasattr(self, 'fig'):
            plt.close(self.fig.number)
            del self.fig

        self.win.destroy()






# Do numeral derivative error propagation
# Derek Fujimoto
# Nov 2021

import numpy as np
import jax
jax.config.update('jax_platform_name', 'cpu')

def get_derror(fn, par, err):
    """
        Propagate errors through a function with the Monte Carlo method
        
        fn: function handle with prototype fn(*par), returns scalar
            note: the function must call jax.numpy functions, not numpy functions
        par: list of parameters passed to fn
        err: list of errors for each parameter (same length and order)
    """
    
    # get gradient functions
    grad_fns = [jax.grad(fn, i) for i in range(len(par))]
    
    # evaluate gradients
    par = np.array(par, dtype=float)
    grad_val = np.array([g(*par) for g in grad_fns])
    
    # calculate error
    err = np.sqrt(np.sum((grad_val*err)**2))
    
    return err



# test inspect tab
# Derek Fujimoto
# Feb 2021

from numpy.testing import *
import numpy as np
from bfit.gui.calculator_nmr_B1 import calculator_nmr_B1

def test_calculate_field():
    calc = calculator_nmr_B1(True)
    
    # voltage
    calc.volt.set('1')
    calc.entry_voltage.focus_set()
    calc.calculate()
    assert_almost_equal(float(calc.field.get()), 0.000960, 
                        err_msg = "calculator nmr B1 set volt = 1")
    calc.root.destroy()

def test_calculate_voltage():
    calc = calculator_nmr_B1(True)
    
    # field
    calc.field.set('1')
    calc.entry_field.focus_set()
    calc.calculate()
    assert_almost_equal(float(calc.volt.get()), 1042.171717, 
                        err_msg = "calculator nmr B1 set field = 1")
    calc.root.destroy()



# Wrapper for easy usage of the iminuit Minuit object
# Derek Fujimoto
# Nov 2020

import inspect
import numpy as np
from iminuit import Minuit
from bfit.fitting.leastsquares import LeastSquares

class minuit(Minuit):
    """
        Chi squared minimization for function of best fit using iminuit
    """
    
    # ====================================================================== #
    def __init__(self, fn, x, y, dy=None, dx=None, dy_low=None, dx_low=None, 
                 fn_prime=None, fn_prime_dx=1e-6, name=None, start=None, 
                 error=None, limit=None, fix=None, print_level=1, **kwargs):
        """
            fn: function handle. f(x, a, b, c, ...)
            x:              x data
            y:              y data
            dy:             error in y
            dx:             error in x
            dy_low:         Optional, if error in y is asymmetric. If not none, dy is upper error
            dx_low:         Optional, if error in y is asymmetric. If not none, dx is upper error
            fn_prime:       Optional, function handle for the first derivative of fn. f'(x, a, b, c, ...)
            fn_prime_dx:    Spacing in x to calculate the derivative for default calculation
            name:           Optional sequence of strings. If set, use this for setting parameter names
            start:          Optional sequence of numbers. Required if the 
                                function takes an array as input or if it has 
                                the form f(x, *pars), and name is not defined. 
                                Default: 1, broadcasted to all inputs
            error           Optional sequence of numbers. Initial step sizes. 
            limit           Optional sequence of limits that restrict the range 
                                format: [[low, high], [low, high], ...]
                                in which a parameter is varied by minuit. 
                                with None, +/- inf used to disable limit
            fix             Optional sequence of booleans. Default: False
            print_level Set the print_level
                            0 is quiet. 
                            1 print out at the end of MIGRAD/HESSE/MINOS. 
                            2 prints debug messages.
                
            kwargs:         passed to Minuit.from_array_func.
                
                To set for parameter "a" one can also assign the following 
                keywords instead of the array inputs
                    
                    a = initial_value
                    error_a = start_error
                    limit_a = (low, high)
                    fix_a = True
        """
        
        # construct least squares object
        ls = LeastSquares(fn = fn, 
                        x = x, 
                        y = y, 
                        dy = dy, 
                        dx = dx, 
                        dy_low = dy_low, 
                        dx_low = dx_low, 
                        fn_prime = fn_prime, 
                        fn_prime_dx = fn_prime_dx)
        self.ls = ls

        # get number of data points
        self.npts = len(x)

        # detect function names
        if name is None:
            
            # get names from code
            name = inspect.getfullargspec(fn).args
            
            # remove self
            if 'self' in name: name.remove('self')
            
            # remove data input
            name = name[1:]
            
            # check for starting parameters
            if start is not None:
                
                # if they don't match assume array input to function
                try:
                    if len(start) != len(name):
                        name = ['x%d'%d for d in range(len(start))]
                
                # start is a scalar to be broadcasted
                except TypeError:
                    pass
                
            # check for bad input
            else:
                for n in name:
                    if '*' in n:
                        raise RuntimeError("If array input must define name or start")
                  
        # set starting values, limits, fixed, errors
        error, kwargs = self._set_start(array=error, 
                                        namestr='error_', 
                                        name=name, 
                                        kwargs=kwargs, 
                                        default=1)
                                        
        limit, kwargs = self._set_start(array=limit, 
                                        namestr='limit_', 
                                        name=name, 
                                        kwargs=kwargs, 
                                        default=[-np.inf, np.inf])
        
        fix, kwargs = self._set_start(  array=fix, 
                                        namestr='fix_', 
                                        name=name, 
                                        kwargs=kwargs, 
                                        default=False)
        
        # are there starting values, limits, fixed, errors?
        is_start = start is not None
        is_error = error is not None
        is_limit = limit is not None
        is_fix   = fix   is not None
        
        keys = kwargs.keys()
        
        # check limit depth 
        if is_start:    
            broadcast_start = get_depth(start) < 1
        
        # iterate parameter names
        for n in name:
            
            # index of name 
            nidx = list(name).index(n)
            
            # start
            if n not in keys:
                
                if is_start:
                    
                    if broadcast_start: kwargs[n] = start
                    else:               kwargs[n] = start[nidx]
                        
                else:
                    kwargs[n] = 1
                
        # make minuit object
        super().__init__(ls, 
                         name = name, 
                         **kwargs)
        
        # set errors, limits, fix
        if is_error:
            self.errors = error
        if is_limit:
            self.limits = limit
        if is_fix:
            self.fixed = fix
                        
        # set errordef for least squares minimization
        self.errordef = 1
        
        # set print level
        self.print_level = print_level
        
    # ====================================================================== #
    def _set_start(self, array, namestr, name, kwargs, default):
        """
            Set starting values, or broadcast if needed
        """
        
        depth = np.array(default).size
        
        # check kwargs for needed values
        items = list(kwargs.items())
        for key, value in items:
            if namestr in key:
                
                idx = list(name).index(key.split('_')[1])
                
                # make new array
                if array is None:   
                    array = [default]*len(name)
                    
                # try to assign if not list, broadcast
                if get_depth(array) < depth:
                    array = [array]*len(name)
                array[idx] = value
                    
                # remove value from kwargs
                del kwargs[key]
                
        return (array, kwargs)
        
    # ====================================================================== #
    @property
    def chi2(self):
        """
            Get the chi2 value
        """
        nfixed = sum(self.fixed)
        narg = len(self.values)
        dof = self.npts - narg + nfixed
        
        if dof <= 0:
            return np.nan
        else:
            return self.fval/dof
    
    # ====================================================================== #
    def get_merrors(self, attribute):
        """
            Get attributes from self.merrors as an array
            
            Valid attributes:
                at_lower_limit
                at_lower_max_fcn
                at upper_limit
                at_upper_max_fcn
                is_valid
                lower
                lower_new_min
                lower_valid
                min
                name
                nfcn
                number
                upper
                upper_new_min
                upper_valid
        """
        
        keylist = tuple(self.merrors.keys())
        
        # check attribute
        if attribute not in self.merrors[keylist[0]].__slots__:
            raise AttributeError('Attribute "%s" not found. Must be one of %s.'%\
                    (attribute, self.merrors[keylist[0]].__slots__))
        
        # get
        return np.array([getattr(self.merrors[k], attribute) for k in keylist])
        
    @property
    def mat_lower_limit(self): return self.get_merrors('at_lower_limit')
    @property
    def mat_lower_max_fcn(self): return self.get_merrors('at_lower_max_fcn')
    @property
    def mat_upper_limit(self): return self.get_merrors('at_upper_limit')
    @property
    def mat_upper_max_fcn(self): return self.get_merrors('at_upper_max_fcn')
    @property
    def mis_valid(self): return self.get_merrors('is_valid')
    @property
    def mlower(self): return self.get_merrors('lower')
    @property
    def mlower_new_min(self): return self.get_merrors('lower_new_min')
    @property
    def mlower_valid(self): return self.get_merrors('lower_valid')
    @property
    def mmin(self): return self.get_merrors('min')
    @property
    def mname(self): return self.get_merrors('name')
    @property
    def mnfcn(self): return self.get_merrors('nfcn')
    @property
    def mnumber(self): return self.get_merrors('number')
    @property
    def mupper(self): return self.get_merrors('upper')
    @property
    def mupper_new_min(self): return self.get_merrors('upper_new_min')
    @property
    def mupper_valid(self): return self.get_merrors('upper_valid')
        
def get_depth(lst, depth=0):
    try: 
        lst[0]
    except (IndexError, TypeError):
        return depth
    else:
        return get_depth(lst[0], depth+1)
        
        
        
        



__all__=['functions', 
         'global_fitter', 
         'global_bdata_fitter', 
         'PulsedFns', 
         'pulsed', 
         'continuous', 
         'fitter', 
         'fitter_curve_fit', 
         'fit_bdata', 
         'decay_31mg', 
         'leastsquares', 
         'minuit',
         '_scipy_misc',
]



# Fit set of data globally 
# Derek Fujimoto
# Nov 2018

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
import os
import time
from bfit.fitting.minuit import minuit
from collections.abc import Iterable
from bfit.fitting.leastsquares import LeastSquares
import warnings

__doc__=\
"""
    Global fitter. 
    
    Uses scipy.optimize.curve_fit to fit a function or list of functions to a 
    set of data with shared parameters.
    
    Usage: 
        
        Construct fitter:
            
            g = global_fitter(x, y, dy, fn, sharelist, npar=-1)
            
            %s
    
        Fit
            g.fit(**fitargs)
        
            %s
            
        Get chi squared
            g.get_chi()
            
            %s
        
        Get fit parameters
            g.get_par()
            
            %s
            
        Draw the result
        
            g.draw(mode='stack', xlabel='', ylabel='', do_legend=False, labels=None, 
                   savefig='', **errorbar_args)
           
            %s
"""

# =========================================================================== #
class global_fitter(object):
    """
        Set up fitting a set of data with arbitrary function. Arbitrary 
        globally shared parameters. 
        
        Instance Variables: 
            
            chi_glbl                global chisqured, also accessible through gchi2
            chi                     list of chisquared values for each data set, also accessible through chi2
            
            fn                      list of fitting function handles
            fixed                   list of fixed variables (corresponds to input)
            fprime_dx               x spacing in calculating centered differences derivative
            
            metadata                array of additional inputs, fixed for each data set
                                    (if len(shared) < len(actual inputs))
            
            minuit                  Minuit object for minimizing with migrad algorithm
            minimizer               string: one of "curve_fit" or "migrad"
            
            npar                    number of parameters in input function
            nsets                   number of data sets
            
            par                     fit results with unnecessary variables stripped 
            par_runwise             fit results run-by-run with all needed inputs, also accessible via values
            std_l, std_u            lower/upper errors with unnecessary variables stripped 
            std_l_runwise           lower errors run-by-run with all needed inputs, also accessible via lower
            std_u_runwise           upper errors run-by-run with all needed inputs, also accessible via upper
            cov                     fit covarince matrix with unnecessary variables stripped, , also accessible via covariance
            cov_runwise             fit covarince matrix run-by-run with all needed inputs
            
            shared                  array of bool of len = nparameters, share parameter if true. 
            sharing_links           2D array of ints, linking global inputs to function-wise inputs
            
            xcat                    concatenated xdata for global fitting
            ycat                    concatenated ydata for global fitting
            dycat                   concatenated dydata for global fitting
            dxcat                   concatenated dydata for global fitting
            dycat_low               concatenated dydata for global fitting
            dxcat_low               concatenated dydata for global fitting
            
            x                       input array of x data sets [array1, array2, ...]
            y                       input array of y data sets [array1, array2, ...]
            dy                      input array of y error sets [array1, array2, ...]
            dx                      input array of x error sets [array1, array2, ...]
            dy_low                  input array of y error sets [array1, array2, ...]
            dx_low                  input array of x error sets [array1, array2, ...]
    """
    
    # class variables
    draw_modes = ('stack', 's', 'new', 'n', 'append', 'a')   # for checking modes
    ndraw_pts = 500             # number of points to draw fits with
    
    # ======================================================================= #
    def __init__(self, fn, x, y, dy=None, dx=None, dy_low=None, dx_low=None, 
                shared=None, fixed=None, metadata=None, fprime_dx=1e-6):
        """
            fn:         function handle OR list of function handles. 
                        MUST specify inputs explicitly if list must have that 
                        len(fn) = len(x) and all function must have the same 
                        inputs in the same order
        
            x, y:       2-list of data sets of equal length. 
                        fmt: [[a1, a2, ...], [b1, b2, ...], ...]
            
            dx, dy:     list of errors in x or y with same format as x and y
            dx_low, dy_low: list of low bound errors in x or y with same format 
                        as x and y. 
                            
                        If not None, then assume dy and dx are the upper errors
                        If None, assume symmetric errors
            
            shared:     tuple of booleans indicating which values to share. 
                        len = number of parameters 
                        
            fixed:      list of booleans indicating if the paramter is to be 
                        fixed to p0 value (same length as p0). Returns best 
                        parameters in order presented, with the fixed 
                        parameters omitted.
                        
            metadata:   array of values to pass to fn which are fixed for each 
                        data set (ex: the temperature of each data set). 
                        format: len(metadata) = number of data sets. 
                        
                        number of parameters is set by len(shared), all 
                        remaining inputs are expected to be metadata inputs
                        function call: fn[i](x[i], *par[i], *metadata[i])
            
            fprime_dx:  x spacing in calculating centered differences derivative
            
        """
        # ---------------------------------------------------------------------
        # Check and assign inputs
        self.fprime_dx = fprime_dx
    
        if (dy is None and dy_low is not None) or (dx is None and dx_low is not None): 
            raise RuntimeError("If specifying lower errors, must also specify dy or dx")
            
        # data types: check and assign
        msg = 'Lengths of input data arrays do not match.\nnsets: %s, %s =  %d, %d\n'
        x = list(x)
        y = list(y)
        
        if not len(x) == len(y):
            raise RuntimeError(msg % ('x', 'y', len(x), len(y)))
        
        self.nsets = len(x)
        
        if dy is not None:          
            dy = list(dy)
            if not self.nsets == len(dy):
                raise RuntimeError(msg % ('n', 'dy', self.nsets, len(dy)))
        
        if dx is not None:          
            dx = list(dx)
            if not self.nsets == len(dx):
                raise RuntimeError(msg % ('n', 'dx', self.nsets, len(dx)))
            
        if dy_low is not None:      
            dy_low = list(dy_low)
            if not self.nsets == len(dy_low):
                raise RuntimeError(msg % ('n', 'dy_low', self.nsets, len(dy_low)))
        
        if dx_low is not None:      
            dx_low = list(dx_low)
            if not self.nsets == len(dx_low):
                raise RuntimeError(msg % ('n', 'dx_low', self.nsets, len(dx_low)))
        
        # shared parameters
        self.shared = np.asarray(shared)
        
        # get number of input parameters
        if shared is None:
            raise RuntimeError("Must specified shared parameter as boolean iterable")
        self.npar = len(shared)

        # if errors, remove points with zero error from data
        if dy is not None or dx is not None:
            for i in range(self.nsets):
                
                # get indexing
                idx = np.full(len(x[i]), False)
                if dy is not None:      idx = idx | (dy[i] != 0)
                if dx is not None:      idx = idx | (dx[i] != 0)
                if dy_low is not None:  idx = idx | (dy_low[i] != 0)
                if dx_low is not None:  idx = idx | (dx_low[i] != 0)
                
                # crop
                x[i] = x[i][idx]
                y[i] = y[i][idx]
                
                if dy is not None:      dy[i] = dy[i][idx]
                if dx is not None:      dx[i] = dx[i][idx]
                if dy_low is not None:  dy_low[i] = dy_low[i][idx]
                if dx_low is not None:  dx_low[i] = dx_low[i][idx]

        # check if list of functions given 
        if not isinstance(fn, Iterable):
            fn = [fn]*self.nsets
        
        # check metadata
        if metadata is not None:
            
            # check for not enough data
            if len(metadata) != self.nsets:
                raise RuntimeError('metadata has the wrong shape: len(metadata) = len(x)')  
                
            # check for 1D input
            metadata = np.asarray(metadata)
            if len(metadata.shape) == 1:
                metadata = np.asarray([metadata]).T
        else:
            metadata = [[]]*self.nsets
        self.metadata = metadata
        
        # expand fixed
        if fixed is not None: 
            if len(np.asarray(fixed).shape) == 1:
                fixed = np.full((self.nsets, self.npar), fixed)
            else:
                fixed = np.asarray(fixed)
        else:
            fixed = np.zeros((self.nsets, self.npar)).astype(bool)
        
        fixed_flat = np.concatenate(fixed)
        
        # check that no shared parameters are fixed
        shared_as_int = self.shared.astype(int)
        for fix in fixed: 
            if any(fix+shared_as_int>1):
                raise RuntimeError('Cannot fix a shared parameter') 
        
        # ---------------------------------------------------------------------
        # Build fitting functions
        
        # get linking indexes    
        sharing_links = [] # [input index] organized by output index position
        p_index = 0
        par_numbers_shared = np.arange(self.npar)
        
        for i in range(self.nsets):
            
            # parameter intdexes
            par_numbers = par_numbers_shared+p_index
            
            # link shared variables
            par_numbers[self.shared] = par_numbers_shared[self.shared]  
            
            # link fixed variables
            par_numbers[fixed[i]] = -par_numbers[fixed[i]]-1
            sharing_links.append(par_numbers)
            
            # iterate
            p_index += self.npar
        
        # reduce too-high indexes
        sharing_links = np.array(sharing_links)        
        unq = np.unique(sharing_links)
        unq = unq[unq>=0]
        for i, u in enumerate(unq):
            sharing_links[sharing_links==u] = i
        
        # save results
        self.fn = fn
        self.x = x
        self.y = y
        self.dy = dy
        self.dx = dx
        self.dy_low = dy_low
        self.dx_low = dx_low
        self.fixed = fixed
        self.sharing_links = sharing_links
        
        # get concatenated data
        self.xcat = np.concatenate(x)
        self.ycat = np.concatenate(y)
        
        if dy is not None:          self.dycat = np.concatenate(dy)
        else:                       self.dycat = None
        
        if dx is not None:          self.dxcat = np.concatenate(dx)
        else:                       self.dxcat = None
        
        if dy_low is not None:      self.dycat_low = np.concatenate(dy_low)
        else:                       self.dycat_low = None
        
        if dx_low is not None:      self.dxcat_low = np.concatenate(dx_low)
        else:                       self.dxcat_low = None
        
    # ======================================================================= #
    def _do_curve_fit(self, master_fn, p0_first, **fitargs):
        """
            Run curve_fit minimmizer
        """
        
        dycat = self.dycat
        absolute_sigma = self.dycat is not None
            
        if self.dxcat is not None:
            warnings.warn("curve_fit minimizer does not account for x errors")
            
        par, cov = curve_fit(master_fn, 
                            self.xcat, 
                            self.ycat, 
                            sigma=dycat, 
                            absolute_sigma=absolute_sigma, 
                            p0 = p0_first, 
                            **fitargs)
                            
        std = np.diag(cov)**0.5
        return (par, std, std, cov)
    
    # ======================================================================= #
    def _do_migrad(self, master_fn, master_fnprime, do_minos, p0_first, **fitargs):
                
        # set args
        limit = fitargs.get('bounds', None)
        if limit is not None:
            limit = np.array(limit).T
        
        kwargs_minuit = {'start':p0_first, 
                         'limit':limit, 
                         'print_level':fitargs.get('print_level', 1), 
                        }
        
        # get minuit obj
        m = minuit(master_fn, self.xcat, self.ycat, 
                            dy = self.dycat, 
                            dx = self.dxcat, 
                            dy_low = self.dycat_low, 
                            dx_low = self.dxcat_low, 
                            fn_prime = master_fnprime,
                            **kwargs_minuit)

        self.ls = m.ls
        self.minuit = m
        
        # minimize
        try:
            m.migrad()    
        except UnicodeEncodeError:  # can't print on older machines
            pass
        
        # check minimum
        if not m.fmin.is_valid:
            raise RuntimeError('Minuit failed to converge to a valid minimum')
        
        # get errors
        if do_minos:
            try:
                m.minos()    
            except UnicodeEncodeError:  # can't print on older machines
                pass
                
            lower = np.abs(m.mlower)
            upper = m.mupper
        else:
            try:
                m.hesse()
            except UnicodeEncodeError:  # can't print on older machines
                pass
        
            lower = m.errors
            upper = lower
        
        # get output
        par = m.values
        cov = m.covariance
        
        return (par, lower, upper, cov)
        
    # ======================================================================= #
    def draw(self, mode='stack', xlabel='', ylabel='', do_legend=False, labels=None, 
             savefig='', **errorbar_args):
        """
            Draw data and fit results. 
            
            mode:           drawing mode. 
                            one of 'stack', 'new', 'append' (or first character 
                                for shorhand)
            
            xlabel/ylabel:  string for axis labels
            
            do_legend:      if true set legend
            
            labels:         list of string to label data
            
            savefig:        if not '', save figure with this name
            
            errorbar_args:  arguments to pass on to plt.errorbar
            
            Returns list of matplotlib figure objects
        """
        
        fig_list = []
        last = 0
        
        # check input
        if mode not in self.draw_modes:
            raise RuntimeError('Drawing mode %s not recognized' % mode)
        
        # get label
        if labels is None:
            labels = ['_no_label_' for i in range(self.nsets)]
        
        # draw all
        for i in range(self.nsets):
            
            # get data
            x, y, = self.x[i], self.y[i]
            
            if self.dy is not None:         dy = self.dy[i]
            else:                           dy = None
            
            if self.dx is not None:         dx = self.dx[i]
            else:                           dx = None
            
            if self.dy_low is not None:     dy_low = self.dy_low[i]
            else:                           dy_low = dy
            
            if self.dx_low is not None:     dx_low = self.dx_low[i]
            else:                           dx_low = dx
            
            f = self.fn[i]
            md = self.metadata[i]
                        
            # make new figure
            if mode in ['new', 'n']:            
                fig_list.append(plt.figure())
            elif len(fig_list) == 0:
                fig_list.append(plt.figure())
            
            # shift x values
            if mode in ['append', 'a']:
                x_draw = x+last
                last = x_draw[-1]
            else:
                x_draw = x
                
            # draw data
            if dy is None:      dy_draw = None
            else:               dy_draw = (dy_low, dy)    
            
            if dx is None:      dx_draw = None
            else:               dx_draw = (dx_low, dx)    
            
            datplt = plt.errorbar(x_draw, y, yerr=dy_draw, xerr=dx_draw, 
                                  label=labels[i], **errorbar_args)
            
            # get color for fit curve
            if mode in ['stack', 's']:
                color = datplt[0].get_color()
            else:
                color = 'k'
            
            # draw fit
            xfit = np.linspace(min(x), max(x), self.ndraw_pts)
            xdraw = np.linspace(min(x_draw), max(x_draw), self.ndraw_pts)
            plt.plot(xdraw, f(xfit, *self.par_runwise[i], *md), color=color, zorder=10)
        
            # plot elements
            plt.ylabel(ylabel)
            plt.xlabel(xlabel)
            
            if do_legend:       plt.legend(fontsize='x-small')
            if savefig!='':     plt.savefig(savefig)
            
            plt.tight_layout()

        return fig_list
        
    # ======================================================================= #
    def fit(self, minimizer='migrad', **fitargs):
        """
            fitargs: parameters to pass to fitter (scipy.optimize.curve_fit) 
            
            p0:         [(p1, p2, ...), ...] innermost tuple is initial parameters 
                            for each data set, list of tuples for all data sets
                            if not enough sets of inputs, last input is copied 
                            for remaining data sets.
                            
                            p0.shape = (nsets, npars)
                    OR
                        (p1, p2, ...) single tuple to set same initial parameters 
                            for all data sets
            
                            p0.shape = (npars, )
            
            bounds:     [((l1, l2, ...), (h1, h2, ...)), ...] similar to p0, but use 
                            2-tuples instead of the 1-tuples of p0
                        
                            bounds.shape = (nsets, 2, npars)
                        
                    OR
                        ((l1, l2, ...), (h1, h2, ...)) single 2-tuple to set same 
                            bounds for all data sets
                            
                            bounds.shape = (2, npars)
                            
            minimizer:      string. One of "trf", "dogbox", or "migrad" indicating 
                            which code to use to minimize the function
            
            do_minos:       if true, and if minimizer==migrad, then run minos errors
            
            returns (parameters, lower errors, upper errors, covariance matrix)
        """
        
        # get values from self
        fixed = self.fixed
        shared = self.shared
        sharing_links = self.sharing_links
        fn = self.fn
        self.minimizer = minimizer
        
        # set default p0
        if 'p0' in fitargs:
            p0 = np.array(list(fitargs['p0']))
            del fitargs['p0']
            
            # expand p0
            if len(p0.shape) == 1:
                p0 = np.full((self.nsets, self.npar), p0)
        else:
            p0 = np.ones((self.nsets, self.npar))
        
        # for fixed parameters
        p0_flat_inv = np.concatenate(p0)[::-1]
        
        # get flattened p0 values which are not fixed
        p0_first = self._flatten(p0)
        
        # set default bounds
        if 'bounds' in fitargs.keys():
            
            # get bounds
            bounds = fitargs['bounds']
            
            # check bounds depth
            depth0 = get_depth(bounds[0])
            depth1 = get_depth(bounds[1])
            
            # error output
            err = RuntimeError("Bad bounds format")
            
            # case: low bound == const, set for all
            if depth0 == 0:
                
                hi = bounds[1]
                
                # case: upper bound == const, set for all
                if depth1 == 0:
                    lo = bounds[0]
                        
                # case: upper bound == list, set for all
                elif depth1 == 1:
                    lo = [bounds[0]]*len(hi)
                    
                else:
                    raise err
                    
            # case: low bound == list 
            elif depth0 == 1:
                
                # case: upper bound == const, set for all
                if depth1 == 0:
                    lo = bounds[0]
                    hi = [bounds[1]]*len(lo)
                    
                # case: upper bound == list, set for all
                elif depth1 == 1:
                    lo = bounds[0]
                    hi = bounds[1]
            
                # case: upper bound == list, set for each
                elif depth1 == 2:
                    hi = [b[1] for b in bounds]
                    lo = [[b[0]]*len(b[1]) for b in bounds]

                else:
                    raise err    

            # case: low bound == list, set for each
            elif depth0 == 2:
                
                lo = [b[0] for b in bounds]
                
                # case: upper bound == const, set for each
                if depth1 == 1:
                    hi = [[b[1]]*len(b[0]) for b in bounds]
            
                # case: upper bound == list, set for each
                elif depth1 == 2:
                    hi = [b[1] for b in bounds]
                    
                else:
                    raise err
            
            else:
                raise err
                
            # expand bounds
            lo = self._expand_bound_lim(lo)
            hi = self._expand_bound_lim(hi)
            
            # flatten bounds 
            lo = self._flatten(lo)
            hi = self._flatten(hi)
            
            # construct bounds
            bounds = (lo, hi)
            fitargs['bounds'] = bounds
    
        # make the master function 
        x = self.x
        rng = range(self.nsets)
        metadata = self.metadata
        fprime_dx = self.fprime_dx
          
        def master_fn(x_unused, *par):
            inputs = np.take(np.hstack((par, p0_flat_inv)), sharing_links)
            return np.concatenate([fn[i](x[i], *inputs[i], *metadata[i]) for i in rng])
        
        self.master_fn = master_fn
            
        # make derivative of master function
        def master_fnprime(x_unused, *par):
            inputs = np.take(np.hstack((par, p0_flat_inv)), sharing_links)
            xhi = np.concatenate([fn[i](x[i]+fprime_dx/2, *inputs[i], *metadata[i]) for i in rng])
            xlo = np.concatenate([fn[i](x[i]-fprime_dx/2, *inputs[i], *metadata[i]) for i in rng])
            return (xhi-xlo)/fprime_dx
            
        self.master_fnprime = master_fnprime
      
        # do curve_fit
        if minimizer in ('trf', 'dogbox'):
            fitargs['method'] = minimizer
            par, std_l, std_u, cov = self._do_curve_fit(master_fn, p0_first, **fitargs)
        
        # do migrad
        elif minimizer in ('migrad', 'minos'):
            fprime_dx = self.fprime_dx
            self.master_fn = master_fn
                    
            par, std_l, std_u, cov = self._do_migrad(master_fn, 
                                                     master_fnprime, 
                                                     minimizer == 'minos', 
                                                     p0_first, 
                                                     **fitargs)
        else:
            raise RuntimeError("Unrecognized minimizer input '%s'" % minimizer)
        
        # to array
        par = np.asarray(par)
        std_l = np.asarray(std_l)
        std_u = np.asarray(std_u)
        cov = np.asarray(cov)
        
        # inflate parameters
        par_out = np.hstack((par, p0_flat_inv))[sharing_links]
        
        zero = np.zeros(len(p0_flat_inv))
        std_l_out = np.hstack((std_l, zero))[sharing_links]
        std_u_out = np.hstack((std_u, zero))[sharing_links]
        
        # inflate covariance matrix
        cov_out = []
        for lnk in sharing_links:
            
            # init
            cov_run = np.zeros((self.npar, self.npar))
            
            # assign
            for i in range(self.npar):
                for j in range(self.npar):
                    
                    # fixed values
                    if lnk[j] < 0 or lnk[i] < 0: 
                        cov_run[i, j] = np.nan
                    
                    # cov
                    else:
                        cov_run[i, j] = cov[lnk[i], lnk[j]]
            cov_out.append(cov_run)
                    
        # return
        self.par = par
        self.std_l = std_l
        self.std_u = std_u
        self.cov = cov
        self.par_runwise = par_out
        self.std_l_runwise = std_l_out
        self.std_u_runwise = std_u_out
        self.cov_runwise = cov_out
        
        return (par_out, std_l_out, std_u_out, cov_out)
    
    # ======================================================================= #
    def get_chi(self):
        """
            Calculate chisq/DOF, both globally and for each function.
            
            sets self.chi and self.chi_glbl
            
            return (global chi, list of chi for each fn)
        """
        return (self.gchi2, self.chi2)

    @property
    def gchi2(self):
        """
            Get global chisquared / DOF
        """
        dof = len(self.xcat)-len(self.par)
        if dof <= 0:
            raise DivisionByZero("Zero degrees of freedom")
        
        if self.minimizer == 'migrad':
            self.chi_glbl = self.minuit.fval/dof
        else:
            
            ls = LeastSquares(self.master_fn, self.xcat, self.ycat, 
                            dy = self.dycat, 
                            dx = self.dxcat, 
                            dy_low = self.dycat_low, 
                            dx_low = self.dxcat_low, 
                            fn_prime = self.master_fnprime)
            
            self.chi_glbl = ls(*self.par) / dof
        return self.chi_glbl
        
    @property
    def chi2(self):
        """
            Get chisquared / DOF for each fit function
        """
        self.chi = []
        for i in range(self.nsets):
            
            # get data
            x = self.x[i]
            y = self.y[i]
            p = self.par_runwise[i]
            f = self.fn[i]
            fx = self.fixed[i]
            m = self.metadata[i]
            
            dy = None if self.dy is None else self.dy[i]
            dx = None if self.dx is None else self.dx[i]
            dy_low = None if self.dy_low is None else self.dy_low[i]
            dx_low = None if self.dx_low is None else self.dx_low[i]
        
            # calc ls sum
            ls = LeastSquares(  f, x, y, 
                                dy = dy, 
                                dx = dx, 
                                dx_low = dx_low, 
                                dy_low = dy_low, 
                            )
            # get dof
            dof = len(x)-self.npar+sum(fx)
            if dof == 0:
                warnings.warn("Zero degrees of freedom for data set %d, using len(x) as dof" % i)
                dof = len(x)
            
            self.chi.append(ls(*np.concatenate((p, m))) / dof)
        
        self.chi = np.array(self.chi)
            
        return self.chi
        
    # ======================================================================= #
    def get_par(self):
        """
            Fetch fit parameters
            
            return 4-tuple of (par, std_l, std_u, cov) with format
            
            ([data1:[par1, par2, ...], data2:[], ...], 
             [data1:[std_l1, std_l2, ...], data2:[], ...], 
             [data1:[std_u1, std_u2, ...], data2:[], ...], 
             [data1:[cov1, cov2, ...], data2:[], ...])
        """
        return (self.values, 
                self.lower, 
                self.upper, 
                self.covariance)
    
    @property
    def values(self):       return self.par_runwise
    
    @property
    def lower(self):        return np.abs(self.std_l_runwise)
    
    @property
    def upper(self):        return self.std_u_runwise
    
    @property
    def covariance(self):   return self.cov_runwise
    
    # ======================================================================= #
    def _expand_bound_lim(self, lim):
        """
            For various bound input formats expand such that all bounds are 
            defined explicitly. 
        """

        lim = np.asarray(lim)
        
        # single float case
        if not lim.shape:
            return np.full((self.nsets, self.npar), np.full(self.npar, lim))
            
        # list case: probably each variable defined
        if len(lim.shape) == 1 and len(lim) == self.npar:
            return np.full((self.nsets, self.npar), lim)
        
        # list case: probably each data set defined in full
        if lim.shape == (self.nsets, self.npar):
            return lim
        
        # we don't know what's happening
        raise RuntimeError('Unexpected bound size input')
        
    # ======================================================================= #
    def _flatten(self, arr):
        """
            Flatten input array based on sharing and fixing. 
            Use for p0, bounds
        """
    
        fixed = self.fixed
        shared = self.shared
    
        arr2 = list(arr[0][~fixed[0]])
        for i in range(1, len(arr)):
            arr2.extend(arr[i][(~fixed[i])*(~shared)])
        return np.array(arr2)

# =========================================================================== #
def get_depth(lst, _n=0):
    """
        get depth of list of lists using first element
        
        lst: list
        _n:   internal depth input (n=0 means not a list)
    """
    
    try:
        lst[0]
    except (TypeError, IndexError):
        return _n
    else:
        return get_depth(lst[0], _n+1)
    
# Add to module docstring
__doc__ = __doc__ % (global_fitter.__init__.__doc__, 
                     global_fitter.fit.__doc__, 
                     global_fitter.get_chi.__doc__, 
                     global_fitter.get_par.__doc__, 
                     global_fitter.draw.__doc__, 
                     )
